#  Copyright (c) 2023 zfit
import pytest

import zfit


def test_parameter_caching():
    zfit.run.set_graph_mode(False)
    # this is to ensure that we fixed the bug (https://github.com/tensorflow/tensorflow/issues/57365) internally
    import tensorflow as tf
    from zfit import z

    x1 = zfit.Parameter("x1", 2.0)
    x2 = zfit.Parameter("x2", 4.0)

    def one_plus(x):
        return x + 1

    cx1 = zfit.ComposedParameter("cx1", one_plus, dependents=[x1])
    cx2 = zfit.ComposedParameter("cx2", one_plus, dependents=[x2])

    @tf.function(autograph=False)
    def f():
        res = x1 + x2**2 / 2
        return res

    def f2():
        res = cx1 + cx2**2 / 2
        return res

    def grad(param):
        param = zfit.z.math._extract_tfparams(param)

        with tf.GradientTape(watch_accessed_variables=False) as tape:
            tape.watch(param)
            value = f()
        return tf.stack(tape.gradient(value, param))

    def grad2(param):
        param = zfit.z.math._extract_tfparams(param)

        with tf.GradientTape(watch_accessed_variables=False) as tape:
            tape.watch(param)
            value = f2()
        return tf.stack(tape.gradient(value, param))

    jitted_grad = z.function(grad)
    jitted_grad2 = z.function(grad2)
    # jitted_grad = tf.function(grad)

    y1 = grad(x1)
    y1_jit = jitted_grad(x1)
    assert abs(y1 - 1.0) < 1e-5  # because d x1 / dx1 = 1
    assert abs(y1_jit - 1.0) < 1e-5
    y2 = grad(x2)
    y2_jit = jitted_grad(x2)

    assert abs(y2 - 4.0) < 1e-5  # because d / dx x**2/2 = x -> 4
    assert abs(y2_jit - y2) < 1e-5
    assert abs(y2_jit - 4.0) < 1e-5

    # use both parameters
    y = grad([x1, x2])
    y_jit = jitted_grad([x1, x2])

    assert abs(y[0] - 1.0) < 1e-5
    assert abs(y[1] - 4.0) < 1e-5
    assert abs(y_jit[0] - 1.0) < 1e-5
    assert abs(y_jit[1] - 4.0) < 1e-5

    # use both parameters, swap order
    y = grad([x2, x1])
    y_jit = jitted_grad([x2, x1])

    assert pytest.approx(y[0], 1e-5) == 4.0
    assert pytest.approx(y[1], 1e-5) == 1.0
    assert pytest.approx(y_jit[0], 1e-5) == 4.0
    assert pytest.approx(y_jit[1], 1e-5) == 1.0

    # test with composed parameters
    y = grad2([x1, x2])
    y_jit = jitted_grad2([x1, x2])

    assert pytest.approx(y[0], 1e-5) == y_jit[0]
    assert pytest.approx(y[1], 1e-5) == y_jit[1]

    # test with composed parameters, swap order
    y = grad2([x2, x1])
    y_jit = jitted_grad2([x2, x1])

    assert pytest.approx(y[0], 1e-5) == y_jit[0]
    assert pytest.approx(y[1], 1e-5) == y_jit[1]



#  Copyright (c) 2023 zfit
from __future__ import annotations

import warnings

import hist
import numpy as np
import tensorflow as tf
import tensorflow_probability as tfp

import zfit.z.numpy as znp
from zfit import z
from zfit.core.interfaces import ZfitData, ZfitRectBinning, ZfitSpace
from zfit.util.ztyping import XTypeInput


def rect_binning_histogramdd(data: XTypeInput, binning: ZfitRectBinning):
    if isinstance(data, ZfitData):
        data = data.value()
    return histogramdd(sample=data, bins=binning.get_edges())


def histogramdd(sample, bins=10, range=None, weights=None, density=None):
    out_dtype = [tf.float64, tf.float64]
    if isinstance(sample, ZfitData):
        sample = sample.value()
        n_obs = sample.n_obs
    else:
        sample = z.convert_to_tensor(sample)
        n_obs = sample.shape[-1]

    none_tensor = tf.constant("NONE_TENSOR", shape=(), name="none_tensor")
    inputs = [sample, bins, range, weights]
    inputs_cleaned = [inp if inp is not None else none_tensor for inp in inputs]

    def histdd(sample, bins, range, weights):
        kwargs = {"sample": sample, "bins": bins, "range": range, "weights": weights}
        new_kwargs = {}
        for key, value in kwargs.items():
            value = value
            is_empty = value == b"NONE_TENSOR"
            try:
                is_empty = bool(is_empty)
            except (
                ValueError
            ):  # if it's a numpy array we need the "all" method, otherwise it's ambiguous
                is_empty = is_empty.all()

            if is_empty:
                value = None

            new_kwargs[key] = value
        return np.histogramdd(**new_kwargs, density=density)

    bincounts, *edges = tf.numpy_function(
        func=histdd, inp=inputs_cleaned, Tout=out_dtype
    )
    bincounts.set_shape(shape=(None,) * n_obs)
    # edges = [edge.set_shape(shape=(None)) for edge in edges]
    return bincounts, edges


def unbinned_to_hist_eager_edgesweightsargs(values, *edges_weights):
    """Same as `unbinned_to_hist_eager` but with the edges and weights as positional arguments.

    This is needed to circumvent the limitation of `tf.numpy_function` that only allows
    numpy arrays as positional arguments and not structures of numpy arrays, such as edges are.

    Args:
        values:
        *edges_weights:

    Returns:
    """
    *edges, weights = edges_weights
    return unbinned_to_hist_eager(values, edges, weights=weights)


def unbinned_to_hist_eager(values, edges, weights=None):
    """Convert an unbinned dataset to a binned dataset in eager mode.

    Args:
        values: Unbinned dataset to convert.
        edges: Edges of the bins.
        weights: Event weights.

    Returns:
        binned_data: Binned dataset.
    """
    if weights is not None and weights.shape == () and None in weights:
        weights = None
    binning = [
        hist.axis.Variable(np.reshape(edge, (-1,)), flow=False) for edge in edges
    ]
    h = hist.Hist(*binning, storage=hist.storage.Weight())
    h.fill(*(values[:, i] for i in range(values.shape[1])), weight=weights)

    return znp.array(h.values(flow=False), znp.float64), znp.array(
        h.variances(flow=False), znp.float64
    )


def unbinned_to_binned(data, space, binned_class=None):
    """Convert an unbinned dataset to a binned dataset.

    Args:
        data: Unbinned dataset to convert.
        space: Space to bin the data in.
        binned_class: Class to use for the binned dataset. Defaults to `BinnedData`.

    Returns:
        binned_data: Binned dataset of type `binned_class`.
    """
    if binned_class is None:
        from zfit._data.binneddatav1 import BinnedData

        binned_class = BinnedData
    if not isinstance(space, ZfitSpace):
        try:
            space = data.space.with_binning(space)
        except Exception as error:
            raise ValueError(
                f"The space provided is not a valid space for the data. "
                f"Either provide a valid space or a binning. "
                f"The error was: {error}"
            ) from error

    values = data.value()
    weights = data.weights
    if weights is not None:
        weights = znp.array(weights)
    edges = tuple(space.binning.edges)
    values, variances = tf.numpy_function(
        unbinned_to_hist_eager_edgesweightsargs,
        inp=[values, *edges, weights],
        Tout=[tf.float64, tf.float64],
    )
    binned = binned_class.from_tensor(space=space, values=values, variances=variances)
    return binned


def unbinned_to_binindex(data, space, flow=False):
    """Calculate the bin index of each data point.

    Args:
        data: Data to calculate the bin index for.
        space: Defines the binning.
        flow: Whether to include the underflow and overflow bins.

    Returns:
        binindex: Tensor with shape (ndata, n_obs) holding the bin index of each data point.
    """
    if flow:
        warnings.warn(
            "Flow currently not fully supported. Values outside the edges are all 0."
        )
    values = [znp.reshape(data.value(ob), (-1,)) for ob in space.obs]
    edges = [znp.reshape(edge, (-1,)) for edge in space.binning.edges]
    bins = [tfp.stats.find_bins(x=val, edges=edge) for val, edge in zip(values, edges)]
    stacked_bins = znp.stack(bins, axis=-1)
    if flow:
        stacked_bins += 1
        bin_is_nan = tf.math.is_nan(stacked_bins)
        zeros = znp.zeros_like(stacked_bins)
        binindices = znp.where(bin_is_nan, zeros, stacked_bins)
        stacked_bins = znp.asarray(binindices, dtype=znp.int32)
    return stacked_bins


def midpoints_from_hist(bincounts, edges):  # TODO: implement correctly, old
    """Calculate the midpoints of a hist and return the non-zero entries, non-zero bincounts and indices.

    Args:
        bincounts: Tensor with shape (nbins_0, ..., nbins_n) with n being the dimension.
        edges: Tensor with shape (n_obs, nbins + 1) holding the position of the edges, assuming a rectangular grid.
    Returns:
        bincounts: the bincounts that are non-zero in a 1-D array corresponding to the indices and the midpoints
        midpoints: the coordinates of the midpoint of each bin with shape (nbincounts, n_obs)
        indices: original position in the bincounts from the input
    """
    bincounts = z.convert_to_tensor(bincounts)
    edges = z.convert_to_tensor(edges)

    midpoints = (edges[:, :-1] + edges[:, 1:]) / 2.0
    midpoints_grid = tf.stack(
        tf.meshgrid(*tf.unstack(midpoints), indexing="ij"), axis=-1
    )
    bincounts_nonzero_index = tf.where(bincounts)
    bincounts_nonzero = tf.gather_nd(bincounts, indices=bincounts_nonzero_index)
    midpoints_nonzero = tf.gather_nd(midpoints_grid, indices=bincounts_nonzero_index)
    return bincounts_nonzero, midpoints_nonzero, bincounts_nonzero_index



#!/usr/bin/env python

#  Copyright (c) 2023 zfit

#
#
# zfit documentation build configuration file, created by
# sphinx-quickstart on Fri Jun  9 13:47:02 2017.
import atexit
import shutil
import sys
from pathlib import Path

import pygit2 as pygit2
import yaml

import zfit

project_dir = Path(__file__).parents[1]
sys.path.insert(0, str(project_dir))

# -- General configuration ---------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
needs_sphinx = "3.0.0"

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx_autodoc_typehints",
    # sphinx_autodoc_typehints must be imported after napoleon to properly work.
    # See https://github.com/agronholm/sphinx-autodoc-typehints/issues/15
    "jupyter_sphinx",
    "sphinx.ext.autosummary",
    "sphinx.ext.viewcode",
    "sphinx.ext.intersphinx",
    "sphinxcontrib.images",
    "sphinx.ext.coverage",
    "sphinx.ext.mathjax",
    "sphinxcontrib.bibtex",
    "sphinx.ext.githubpages",
    "sphinx.ext.todo",
    "sphinx_copybutton",
    "sphinxcontrib.youtube",
    "sphinx_panels",
    "seed_intersphinx_mapping",
    "myst_nb",
    "sphinx_togglebutton",
]

panels_add_bootstrap_css = (
    False  # for sphinx_panel, use custom css from theme, not bootstrap
)

# releases_github_path = "zfit/zfit"  # TODO: use releases or similar?
# releases_document_name = "../CHANGELOG.rst"

# nb_execution_mode = "force"  # use if needed and cache should be ignored
nb_execution_mode = "cache"
if nb_execution_mode == "cache":
    jupyter_cache_path = project_dir.joinpath("docs", ".cache", "myst-nb")
    jupyter_cache_path.mkdir(parents=True, exist_ok=True)
    nb_execution_cache_path = str(jupyter_cache_path)

source_suffix = {
    ".ipynb": "myst-nb",
    ".myst": "myst-nb",
}
myst_enable_extensions = [
    "amsmath",
    "colon_fence",
    "deflist",
    "dollarmath",
    "html_image",
]

bibtex_bibfiles = [str(project_dir.joinpath("docs", "refs.bib"))]

zfit_tutorials_path = project_dir.joinpath("docs", "_tmp", "zfit-tutorials")
atexit.register(lambda path=zfit_tutorials_path: shutil.rmtree(path))
pygit2.clone_repository("https://github.com/zfit/zfit-tutorials", zfit_tutorials_path)

zfit_images_path = project_dir.joinpath("docs", "images")
docs_images_path = project_dir.joinpath("docs", "_static", "images")
docs_images_path.mkdir(parents=True, exist_ok=True)
atexit.register(lambda path=docs_images_path: shutil.rmtree(path))
shutil.copytree(zfit_images_path, docs_images_path, dirs_exist_ok=True)

nb_execution_in_temp = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "zfit"
copyright = zfit.__copyright__
author = zfit.__author__

# The version info for the project you're documenting, acts as replacement
# for |version| and |release|, also used in various other places throughout
# the built documents.
#
# The short X.Y version.
version = zfit.__version__.split("+")[0]
# The full version, including alpha/beta/rc tags.
release = zfit.__version__

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# Automatically add substitutions to all RST files.
with open("subst_types.txt") as subst_types:
    rst_epilog = subst_types.read()

# add whitespaces to the internal commands. Maybe move to preprocessing?
rst_epilog += """
.. |wzw| unicode:: U+200B
   :trim:

"""
# ..  replace:: |wzw|
#
# .. |@docend| replace:: |wzw|
# """
with open(project_dir / "utils/api/argdocs.yaml") as replfile:
    replacements = yaml.load(replfile, Loader=yaml.Loader)
for replacement_key in replacements:
    rst_epilog += f"""
.. |@doc:{replacement_key}| replace:: |wzw|

.. |@docend:{replacement_key}| replace:: |wzw|
"""
with open("hyperlinks.txt") as hyperlinks:
    rst_epilog += hyperlinks.read()

# makes the jupyter extension executable
jupyter_sphinx_thebelab_config = {
    # 'bootstrap': False,
    # 'requestKernel': True,
    "binderOptions": {
        "repo": "zfit/zfit-tutorials",
        "binderUrl": "https://mybinder.org",
        "repoProvider": "github",
    },
}

html_favicon = "images/zfit-favicon.png"

# -- Napoleon settings ---------------------------------------------

using_numpy_style = False  # False -> google style
napoleon_google_docstring = not using_numpy_style
napoleon_numpy_docstring = using_numpy_style
napoleon_include_init_with_doc = False
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = True
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = False
napoleon_use_param = True
napoleon_use_rtype = True

# -- sphinx_autodoc_typehints settings ---------------------------------------------

# if True, set typing.TYPE_CHECKING to True to enable “expensive” typing imports
set_type_checking_flag = True
# if True, class names are always fully qualified (e.g. module.for.Class). If False, just the class
# name displays (e.g. Class)
typehints_fully_qualified = False
# (default: False): If False, do not add ktype info for undocumented parameters. If True, add stub documentation for
# undocumented parameters to be able to add type info.
always_document_param_types = False
# (default: True): If False, never add an :rtype: directive. If True, add the :rtype: directive if no existing :rtype:
# is found.
typehints_document_rtype = True

# -- autodoc settings ---------------------------------------------

# also doc __init__ docstrings
autoclass_content = "both"
autodoc_member_order = "bysource"
autodoc_default_options = {
    "show-inheritance": True,
    "inherited-members": True,
}
autodoc_inherit_docstrings = False

# -- autosummary settings ---------------------------------------------

autosummary_generate = True
autosummary_generate_overwrite = True

# -- sphinx.ext.todo settings ---------------------------------------------
# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True

# -- Options for HTML output -------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#

html_theme = "pydata_sphinx_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# These paths are either relative to html_static_path
# or fully qualified paths (eg. https://...)
html_css_files = [
    "css/custom.css",
]

# Theme options are theme-specific and customize the look and feel of a
# theme further.

html_theme_options = {
    "logo": {
        "image_light": "images/zfit-logo_400x168.png",
        "image_dark": "images/zfit-logo-light_400x168.png",
    },
    "github_url": "https://github.com/zfit/zfit",
    "use_edit_page_button": True,
    "navigation_depth": 3,
    "search_bar_text": "Search zfit...",
    "navigation_with_keys": True,
    "search_bar_position": "sidebar",
    "icon_links": [
        {}
    ],  # temporary fix for https://github.com/pydata/pydata-sphinx-theme/issues/1220
    # "repository_url": "https://github.com/zfit/zfit",  # adding jupyter book somehow?
    # "repository_branch": "develop",
    # "path_to_docs": "docs",
}

html_context = {
    "github_user": "zfit",
    "github_repo": "zfit",
    "github_version": "develop",
    "doc_path": "docs",
}

# -- Options for HTMLHelp output ---------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "zfitdoc"

# -- Options for LaTeX output ------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass
# [howto, manual, or own class]).
latex_documents = [
    (master_doc, "zfit.tex", "zfit Documentation", "zfit", "manual"),
]

# -- Options for manual page output ------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "zfit", "zfit Documentation", [author], 1)]

# -- Options for Texinfo output ----------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "zfit",
        "zfit Documentation",
        author,
        "zfit",
        "One line description of project.",
        "Miscellaneous",
    ),
]
# cross reference
default_role = "py:obj"
primary_domain = "py"
# nitpicky = True  # warn if cross-references are missing
# nitpick_ignore = [
#     ("py:class", "tensorflow.keras.losses.Loss"),
# ]
# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    # 'numdifftools': ('https://numdifftools.readthedocs.io/en/latest/index.html', None),
    "tensorflow": (
        "https://www.tensorflow.org/api_docs/python",
        "https://raw.githubusercontent.com/GPflow/tensorflow-intersphinx/master/tf2_py_objects.inv",
    ),
    "tensorflow_probability": (
        "https://www.tensorflow.org/probability/api_docs/python",
        " https://raw.githubusercontent.com/GPflow/tensorflow-intersphinx/master/tfp_py_objects.inv",
    ),
    "uproot": ("https://uproot.readthedocs.io/en/latest/", None),
    "python": ("https://docs.python.org/3", None),
    "matplotlib": ("https://matplotlib.org/stable/", None),
    "pandas": ("https://pandas.pydata.org/docs/", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
}



#  Copyright (c) 2023 zfit
import copy
from contextlib import suppress

import numpy as np
import pytest
from frozendict import frozendict

import zfit

rndgen = np.random.Generator(np.random.PCG64(8213))
rndgens = [np.random.Generator(np.random.PCG64(8213 + i)) for i in range(10)]


def test_serial_space():
    import zfit

    space = zfit.Space("x", (-1, 1))
    json1 = space.to_json()
    space2 = zfit.Space.get_repr().parse_raw(json1).to_orm()
    space2_direct = zfit.Space.hs3.from_json(json1)

    assert space == space2
    assert space == space2_direct
    json2 = space2.to_json()
    assert json1 == json2
    space3 = zfit.Space.get_repr().parse_raw(json2).to_orm()
    assert space2 == space3
    assert space == space3


def test_serial_param():
    import zfit

    param = zfit.Parameter("mu", 0.1, -1, 1)

    json1 = param.to_json()
    param2 = zfit.Parameter.get_repr().parse_raw(json1).to_orm()

    assert param == param2
    json2 = param2.to_json()
    assert json1 == json2
    param3 = zfit.Parameter.get_repr().parse_raw(json2).to_orm()
    assert param2 == param3
    assert param == param3


def gauss(extended=None, **kwargs):
    import zfit

    mu = zfit.Parameter("mu_gauss", 0.1, -1, 1)
    sigma = zfit.Parameter("sigma_gauss", 0.1, 0, 1)
    obs = zfit.Space("obs", (-3, 3))
    return zfit.pdf.Gauss(
        mu=mu, sigma=sigma, obs=obs, extended=extended, name="MyGaussName"
    )


def prod2dgauss(extended=None, **kwargs):
    import zfit

    mu = zfit.Parameter("mu_gauss", 0.1, -1, 1)
    sigma1 = zfit.Parameter("sigma_gauss", 0.1, 0, 1)
    obs1 = zfit.Space("obs1", (-3, 7))
    sigma2 = zfit.Parameter("sigma_gauss2", 0.1, 0, 1)
    obs2 = zfit.Space("obs2", (-13, 5))
    gauss1 = zfit.pdf.Gauss(mu=mu, sigma=sigma1, obs=obs1)
    gauss2 = zfit.pdf.Gauss(mu=mu, sigma=sigma2, obs=obs2)
    prod = zfit.pdf.ProductPDF([gauss1, gauss2], extended=extended, name="prod2dgauss")
    return prod


def cauchy(extended=None, **kwargs):
    import zfit

    m = zfit.Parameter("m_cauchy", 0.1, -1, 1)
    gamma = zfit.Parameter("gamma_cauchy", 0.1, 0, 1)
    obs = zfit.Space("obs", (-3, 3))
    return zfit.pdf.Cauchy(m=m, gamma=gamma, obs=obs, extended=extended)


def exponential(extended=None, **kwargs):
    import zfit

    lam = zfit.Parameter("lambda_exp", 0.1, -1, 1)
    obs = zfit.Space("obs", (-3, 3))
    return zfit.pdf.Exponential(lam=lam, obs=obs, extended=extended)


def crystalball(extended=None, **kwargs):
    import zfit

    alpha = zfit.Parameter("alphacb", 0.1, -1, 1)
    n = zfit.Parameter("ncb", 0.1, 0, 1)
    mu = zfit.Parameter("mucb", 0.1, -1, 1)
    sigma = zfit.Parameter("sigmacb", 0.1, 0, 1)
    obs = zfit.Space("obs", (-3, 3))
    return zfit.pdf.CrystalBall(
        alpha=alpha, n=n, mu=mu, sigma=sigma, obs=obs, extended=extended
    )


def doublecb(extended=None, **kwargs):
    import zfit

    alphaL = zfit.Parameter("alphaL_dcb", 0.1, -1, 1)
    nL = zfit.Parameter("nL_dcb", 0.1, 0, 1)
    alphaR = zfit.Parameter("alphaR_dcb", 0.1, -1, 1)
    nR = zfit.Parameter("nR_dcb", 0.1, 0, 1)
    mu = zfit.Parameter("mu_dcb", 0.1, -1, 1)
    sigma = zfit.Parameter("sigma_dcb", 0.1, 0, 1)
    obs = zfit.Space("obs", (-3, 3))
    return zfit.pdf.DoubleCB(
        alphal=alphaL,
        nl=nL,
        alphar=alphaR,
        nr=nR,
        mu=mu,
        sigma=sigma,
        obs=obs,
        extended=extended,
    )


def legendre(extended=None, **kwargs):
    import zfit

    obs = zfit.Space("obs", (-3, 3))
    coeffs = [zfit.Parameter(f"coeff{i}_legendre", 0.1, -1, 1) for i in range(5)]
    return zfit.pdf.Legendre(obs=obs, coeffs=coeffs, extended=extended)


def chebyshev(extended=None, **kwargs):
    import zfit

    obs = zfit.Space("obs", (-3, 3))
    coeffs = [zfit.Parameter(f"coeff{i}_cheby", 0.1, -1, 1) for i in range(5)]
    return zfit.pdf.Chebyshev(obs=obs, coeffs=coeffs, extended=extended)


def chebyshev2(extended=None, **kwargs):
    import zfit

    obs = zfit.Space("obs", (-3, 3))
    coeffs = [zfit.Parameter(f"coeff{i}_cheby2", 0.1, -1, 1) for i in range(5)]
    return zfit.pdf.Chebyshev2(obs=obs, coeffs=coeffs, extended=extended)


def laguerre(extended=None, **kwargs):
    import zfit

    obs = zfit.Space("obs", (-3, 3))
    coeffs = [zfit.Parameter(f"coeff{i}_laguerre", 0.1) for i in range(5)]
    return zfit.pdf.Laguerre(obs=obs, coeffs=coeffs, extended=extended)


def hermite(extended=None, **kwargs):
    import zfit

    obs = zfit.Space("obs", (-3, 3))
    coeffs = [zfit.Parameter(f"coeff{i}_hermite", 0.1, -1, 1) for i in range(5)]
    return zfit.pdf.Hermite(obs=obs, coeffs=coeffs, extended=extended)


def kde1dimexact(pdfs=None, extended=None, **kwargs):
    data = np.array(
        [
            0.5510963,
            0.46542518,
            -1.3578327,
            -1.1137842,
            0.30369744,
            -0.49643811,
            0.68624974,
            0.2091461,
            -0.32975698,
            0.71040372,
            1.95709914,
            -0.0430226,
            1.28575821,
            -0.37646677,
            -0.25665507,
            1.67085858,
            0.05659317,
            0.52634881,
            0.58759272,
            -0.24556029,
            -0.74070669,
        ]
    )
    import zfit

    obs = zfit.Space("obs", (-3, 3))
    return zfit.pdf.KDE1DimExact(data, obs=obs, extended=extended)


def kde1dgrid(pdfs=None, extended=None, **kwargs):
    data = np.array(
        [
            0.5510963,
            0.46542518,
            -1.3578327,
            -1.1137842,
            0.30369744,
            -0.49643811,
            0.68624974,
            0.2091461,
            -0.32975698,
            0.71040372,
            1.95709914,
            -0.0430226,
            1.28575821,
            -0.37646677,
            -0.25665507,
            1.67085858,
            0.05659317,
            0.52634881,
            0.58759272,
            -0.24556029,
            -0.74070669,
        ]
    )
    import zfit

    obs = zfit.Space("obs", (-3, 3))
    return zfit.pdf.KDE1DimGrid(data, obs=obs, extended=extended, num_grid_points=512)


def kde1dfft(pdfs=None, extended=None, **kwargs):
    data = np.array(
        [
            0.5510963,
            0.46542518,
            -1.3578327,
            -1.1137842,
            0.30369744,
            -0.49643811,
            0.68624974,
            0.2091461,
            -0.32975698,
            0.71040372,
            1.95709914,
            -0.0430226,
            1.28575821,
            -0.37646677,
            -0.25665507,
            1.67085858,
            0.05659317,
            0.52634881,
            0.58759272,
            -0.24556029,
            -0.74070669,
        ]
    )
    import zfit

    obs = zfit.Space("obs", (-3, 3))
    return zfit.pdf.KDE1DimFFT(data=data, obs=obs, extended=extended, weights=data**2)


def kde1disj(pdfs=None, extended=None, **kwargs):
    data = np.array(
        [
            0.5510963,
            0.46542518,
            -1.3578327,
            -1.1137842,
            0.30369744,
            -0.49643811,
            0.68624974,
            0.2091461,
            -0.32975698,
            0.71040372,
            1.95709914,
            -0.0430226,
            1.28575821,
            -0.37646677,
            -0.25665507,
            1.67085858,
            0.05659317,
            0.52634881,
            0.58759272,
            -0.24556029,
            -0.74070669,
        ]
    )
    import zfit

    obs = zfit.Space("obs", (-3, 3))
    data = zfit.data.Data.from_numpy(obs=obs, array=data)
    return zfit.pdf.KDE1DimFFT(data=data, extended=extended)


basic_pdfs = [
    gauss,
    cauchy,
    exponential,
    crystalball,
    doublecb,
    legendre,
    chebyshev,
    chebyshev2,
    laguerre,
    hermite,
    kde1dimexact,
    kde1dgrid,
    kde1dfft,
    kde1disj,
]
basic_pdfs.reverse()


def sumpdf(pdfs=None, fracs=True, extended=None, **kwargs):
    if pdfs is None:
        pdfs = [pdf() for pdf in basic_pdfs]
    import zfit

    if fracs:
        fracs = [zfit.Parameter(f"frac{i}", 0.1, -1, 1) for i in range(len(pdfs))]
    return zfit.pdf.SumPDF(pdfs=pdfs, fracs=fracs, extended=extended)


def productpdf(pdfs=None, extended=None, **kwargs):
    if pdfs is None:
        pdfs = [pdf() for pdf in basic_pdfs[:3]]
    import zfit

    return zfit.pdf.ProductPDF(pdfs=pdfs, extended=extended)


def complicatedpdf(pdfs=None, extended=None, **kwargs):
    import zfit

    basic_pdfs_made = [pdf() for pdf in basic_pdfs]
    pdfs1 = basic_pdfs_made[:3]
    pdfs2 = basic_pdfs_made[3:5]
    pdfs3 = basic_pdfs_made[5:]
    pdfs4 = basic_pdfs_made[1:4]

    sum1 = zfit.pdf.SumPDF(
        pdfs=pdfs1,
        fracs=[
            zfit.Parameter(f"frac_sum1_{i}", 0.1, -1, 1) for i in range(len(pdfs1) - 1)
        ],
    )
    pdfs2.append(sum1)
    sum2 = zfit.pdf.SumPDF(
        pdfs=pdfs2,
        fracs=[
            zfit.Parameter(f"frac_sum2_{i}", 0.1, -1, 1) for i in range(len(pdfs2) - 1)
        ],
    )
    pdfs3.append(sum2)
    prod1 = zfit.pdf.ProductPDF(pdfs=pdfs3)
    pdfs4.append(prod1)
    sum3 = zfit.pdf.SumPDF(
        pdfs=pdfs4,
        extended=extended,
        fracs=[
            zfit.Parameter(f"frac_sum3_{i}", 0.1, -1, 1) for i in range(len(pdfs4) - 1)
        ],
        name="complicatedpdf",
    )
    return sum3


def convolutionpdf(pdf=None, extended=None, **kwargs):
    kernel = gauss()
    func = cauchy()
    import zfit

    return zfit.pdf.FFTConvPDFV1(kernel=kernel, func=func, extended=extended)


def conditionalpdf(pdf=None, extended=None, **kwargs):
    if pdf is None:
        pdf = gauss()
    import zfit

    return zfit.pdf.ConditionalPDFV1(
        pdf=pdf, cond={pdf.params["sigma"]: pdf.space}, extended=extended
    )


all_pdfs = (
    basic_pdfs
    + [
        sumpdf,
        productpdf,
        convolutionpdf,
        complicatedpdf,
    ]
    + [prod2dgauss]
)

all_constraints = [
    lambda: zfit.constraint.GaussianConstraint(
        params=[zfit.Parameter("mu", 0.0, -1, 1), zfit.Parameter("sigma", 1.0, 0, 10)],
        observation=[0.0, 1.0],
        uncertainty=[0.1, 0.5],
    ),
    lambda: zfit.constraint.PoissonConstraint(
        params=[zfit.Parameter("mu", 0.1, -1, 1), zfit.Parameter("sigma", 1.0, 0, 10)],
        observation=[0.1, 1.2],
    ),
    lambda: zfit.constraint.LogNormalConstraint(
        params=[zfit.Parameter("mu", 0.1, -1, 1), zfit.Parameter("sigma", 1.0, 0, 10)],
        observation=[0.1, 1.2],
        uncertainty=[0.1, 0.5],
    ),
]


# TODO(serialization): add to serializer
@pytest.mark.parametrize(
    "ext_Loss",
    [(False, zfit.loss.UnbinnedNLL), (True, zfit.loss.ExtendedUnbinnedNLL)],
    ids=lambda x: x[1].__name__,
)
@pytest.mark.parametrize("pdf_factory", all_pdfs, ids=lambda x: x.__name__)
@pytest.mark.parametrize(
    "Constraint", [None] + all_constraints, ids=lambda x: (x is not None) and x.__name__
)
def test_loss_serialization(ext_Loss, pdf_factory, Constraint, request):
    import zfit

    extended, Loss = ext_Loss
    pdf = pdf_factory(extended=extended)
    assert pdf.is_extended == extended, "Error in testing setup"
    data = zfit.Data.from_numpy(
        obs=pdf.space, array=rndgens[0].normal(size=(1000, pdf.n_obs))
    )
    constraint = Constraint() if Constraint is not None else None
    loss = Loss(pdf, data, constraints=constraint)
    loss_dict = loss.to_dict()
    loss_asdf = loss.to_asdf()

    loss_truth_dumped = pytest.helpers.get_truth(
        f"hs3_loss/{loss.name}",
        f"{loss.name}_{pdf.name}_{'unconstr' if constraint is None else constraint.name}.asdf",
        request,
        newval=loss_asdf,
    )
    loss_truth_tree = loss_truth_dumped.tree
    loss_asdf_tree = loss_asdf.tree
    for l in [loss_asdf_tree, loss_truth_tree]:
        l.pop("asdf_library", None)
        l.pop("history", None)

    # assert loss_asdf_tree == loss_truth_tree
    loss_loaded = loss.from_asdf(loss_asdf)

    print(loss_loaded)


@pytest.mark.parametrize("Constraint", all_constraints, ids=lambda x: x.__name__)
def test_constraint_dumpload(Constraint, request):
    constraint = Constraint()
    constraint_dict = constraint.to_dict()
    constraint_json = constraint.to_json()
    constraint_loaded_truth = pytest.helpers.get_truth(
        "hs3_constraint", f"{constraint.name}.json", request, newval=constraint_json
    )
    constraint_loaded = constraint.__class__.from_json(constraint_json).to_dict()
    constraint_loaded_truth["library"] = "ZFIT_ARBITRARY_VALUE"
    (
        constraint_loaded_cleaned,
        constraint_loaded_truth_cleaned,
    ) = pytest.helpers.cleanup_hs3(constraint_loaded, constraint_loaded_truth)
    assert constraint_loaded_cleaned == constraint_loaded_truth_cleaned


@pytest.mark.parametrize("pdf", all_pdfs, ids=lambda x: x.__name__)
@pytest.mark.parametrize("extended", [True, None], ids=["extended", "not extended"])
def test_serial_hs3_pdfs(pdf, extended):
    import zfit
    import zfit.z.numpy as znp
    import zfit.serialization as zserial

    pdf = pdf(extended=extended)
    scale = zfit.Parameter("yield", 100.0, 0, 1000)
    if extended and not pdf.is_extended:
        pdf.set_yield(scale)

    hs3json = zserial.Serializer.to_hs3(pdf)
    loaded = zserial.Serializer.from_hs3(hs3json)

    loaded_pdf = list(loaded["distributions"].values())[0]
    assert str(pdf) == str(loaded_pdf)
    x = znp.random.uniform(-3, 3, size=(107, pdf.n_obs))
    assert np.allclose(pdf.pdf(x), loaded_pdf.pdf(x))
    if extended:
        scale.set_value(0.6)
        assert np.allclose(pdf.ext_pdf(x), loaded_pdf.ext_pdf(x))


def test_replace_matching():
    import zfit
    import zfit.serialization as zserial

    original_dict = {
        "metadata": {
            "HS3": {"version": "experimental"},
            "serializer": {"lib": "zfit", "version": str(zfit.__version__)},
        },
        "distributions": {
            "Gauss": {
                "mu": {
                    "floating": True,
                    "max": 1.0,
                    "min": -1.0,
                    "name": "mu",
                    "step_size": 0.001,
                    "type": "Parameter",
                    "value": 0.10000000149011612,
                },
                "sigma": {
                    "floating": True,
                    "max": 1.0,
                    "min": 0.0,
                    "name": "sigma",
                    "step_size": 0.001,
                    "type": "Parameter",
                    "value": 0.10000000149011612,
                },
                "type": "Gauss",
                "x": {"max": 3.0, "min": -3.0, "name": "obs", "type": "Space"},
            }
        },
        "variables": {
            "mu": {
                "floating": True,
                "max": 1.0,
                "min": -1.0,
                "name": "mu",
                "step_size": 0.001,
                "type": "Parameter",
                "value": 0.10000000149011612,
            },
            "obs": {"max": 3.0, "min": -3.0, "name": "obs", "type": "Space"},
            "sigma": {
                "floating": True,
                "max": 1.0,
                "min": 0.0,
                "name": "sigma",
                "step_size": 0.001,
                "type": "Parameter",
                "value": 0.10000000149011612,
            },
        },
    }

    target_dict = {
        "metadata": {
            "HS3": {"version": "experimental"},
            "serializer": {"lib": "zfit", "version": str(zfit.__version__)},
        },
        "distributions": {
            "Gauss": {"mu": "mu", "sigma": "sigma", "type": "Gauss", "x": "obs"}
        },
        "variables": {
            "mu": {
                "floating": True,
                "max": 1.0,
                "min": -1.0,
                "name": "mu",
                "step_size": 0.001,
                "type": "Parameter",
                "value": 0.10000000149011612,
            },
            "obs": {"max": 3.0, "min": -3.0, "name": "obs", "type": "Space"},
            "sigma": {
                "floating": True,
                "max": 1.0,
                "min": 0.0,
                "name": "sigma",
                "step_size": 0.001,
                "type": "Parameter",
                "value": 0.10000000149011612,
            },
        },
    }

    parameter = frozendict({"name": None, "min": None, "max": None})
    replace_forward = {parameter: lambda x: x["name"]}
    target_test = copy.deepcopy(original_dict)
    target_test["distributions"] = zserial.serializer.replace_matching(
        target_test["distributions"], replace_forward
    )
    assert target_test == target_dict
    replace_backward = {
        k: lambda x=k: target_dict["variables"][x]
        for k in target_dict["variables"].keys()
    }
    original_test = copy.deepcopy(target_dict)
    original_test["distributions"] = zserial.serializer.replace_matching(
        original_test["distributions"], replace_backward
    )

    assert original_test == original_dict


@pytest.mark.parametrize("pdfcreator", all_pdfs, ids=lambda x: x.__name__)
def test_dumpload_pdf(pdfcreator):
    import zfit.z.numpy as znp

    pdf = pdfcreator()
    params = list(pdf.get_params())
    if len(params) > 0:
        param1 = params[0]
    else:
        param1 = None
    json1 = pdf.to_dict()
    gauss2 = type(pdf).from_dict(json1)
    try:
        json1 = pdf.to_json()
    except zfit.exception.NumpyArrayNotSerializableError:
        pass  # KDEs or similar
    else:
        gauss2 = pdf.get_repr().parse_raw(json1).to_orm()

    assert str(pdf) == str(gauss2)

    json2 = gauss2.to_dict()
    gauss3 = type(pdf).from_dict(json2)
    try:
        json1 = pdf.to_json()
        json2 = gauss2.to_json()
    except zfit.exception.NumpyArrayNotSerializableError:
        pass  # KDEs or similarjson1
    else:
        gauss3 = pdf.get_repr().parse_raw(json2).to_orm()

        json1cleaned = json1
        json2cleaned = json2
        for i in range(1000, 0, -1):
            json2cleaned = json2cleaned.replace(f"autoparam_{i}", "autoparam_ANY")
            json1cleaned = json2cleaned.replace(f"autoparam_{i}", "autoparam_ANY")
        assert json1cleaned == json2cleaned  # Just a technicality

    x = znp.random.uniform(-3, 3, size=(100, pdf.n_obs))
    assert np.allclose(pdf.pdf(x), gauss3.pdf(x))
    assert np.allclose(gauss2.pdf(x), gauss3.pdf(x))
    if param1 is not None:
        param1.set_value(0.6)
    assert np.allclose(pdf.pdf(x), gauss3.pdf(x))
    assert np.allclose(gauss2.pdf(x), gauss3.pdf(x))


param_factories = [
    lambda: zfit.param.ComposedParameter(
        "test", lambda x: x, [zfit.Parameter("x", 1.0, 0.5, 1.4)]
    ),
    lambda: zfit.Parameter("test1", 5.0, 3.0, 10.0),
    lambda: zfit.Parameter("test1", 5.0, 3.0, 10.0, step_size=1.1),
    lambda: zfit.Parameter("test1", 5.0, step_size=1.1),
    lambda: zfit.Parameter("test1", 5.0, 3.0, 10.0, step_size=1.1, floating=False),
    lambda: zfit.Parameter("test1", 5.0, 3.0, 10.0, floating=False),
    lambda: zfit.Parameter("test1", 5.0, floating=True),
    lambda: zfit.param.ConstantParameter("const1", 5.0),
    lambda: zfit.Space("obs", (-3.0, 5.0)),
]


@pytest.mark.parametrize("param_factory", param_factories)
def test_params_dumpload(param_factory):
    param = param_factory()
    json = param.to_json()
    param_loaded = param.get_repr().parse_raw(json).to_orm()
    assert param == param_loaded
    assert json == param_loaded.to_json()


data_factories = enumerate(
    [
        lambda: zfit.data.Data.from_numpy(
            obs=zfit.Space("obs1", (-3.0, 5.0)), array=rndgens[1].normal(size=(100, 1))
        ),
        lambda: zfit.data.Data.from_numpy(
            obs=zfit.Space("obs1", (-3.0, 5.0)),
            array=rndgens[2].normal(size=(100, 1)),
            weights=rndgens[2].normal(size=(100, 1)),
        ),
        lambda: zfit.data.Data.from_numpy(
            obs=zfit.Space("obs1", (-3.0, 5.0)) * zfit.Space("obs2", (-13.0, 15.0)),
            array=rndgens[3].normal(size=(100, 2)),
            weights=rndgens[3].normal(size=(100, 1)),
        ),
        lambda: zfit.data.Data.from_numpy(
            obs=zfit.Space("obs1", (-3.0, 5.0)) * zfit.Space("obs2", (-13.0, 15.0)),
            array=rndgens[4].normal(size=(100, 2)),
        ),
    ]
)


@pytest.mark.parametrize("data_factory", data_factories)
def test_data_dumpload(data_factory, request):
    i, factory = data_factory
    data = factory()
    asdf_dumped = data.to_asdf()
    data_truth_dumped = pytest.helpers.get_truth(
        "hs3_data", f"basic_data_dumpload_{i}.asdf", request, newval=asdf_dumped
    )
    data_loaded1 = data.__class__.from_asdf(
        asdf_dumped
    )  # class just to make sure we don't use the instance method
    asdf_dumped2 = (
        data_loaded1.to_asdf()
    )  # dump and load two times to check consistency
    data_loaded2 = data.__class__.from_asdf(
        asdf_dumped2
    )  # class just to make sure we don't use the instance method
    data_truth_tree = dict(data_truth_dumped.tree)
    data_tree = dict(asdf_dumped.tree)
    data_tree2 = dict(asdf_dumped2.tree)

    for d in [data_truth_tree, data_tree, data_tree2]:
        d.pop("asdf_library", None)
        d.pop("history", None)

    # test and pop all the arrays in the dict, then compare the rest
    true_weigths = data_truth_tree.pop("weights", 1)
    np.testing.assert_allclose(true_weigths, data_tree.pop("weights", 1))
    np.testing.assert_allclose(true_weigths, data_tree2.pop("weights", 1))

    true_data = data_truth_tree.pop("data")
    np.testing.assert_allclose(true_data, data_tree.pop("data"))
    np.testing.assert_allclose(true_data, data_tree2.pop("data"))
    np.testing.assert_allclose(data.value(), data_loaded1.value())
    np.testing.assert_allclose(data.value(), data_loaded2.value())
    assert (data.weights is None) == (data_loaded1.weights is None)
    assert (data.weights is None) == (data_loaded2.weights is None)
    if data.weights is not None:
        np.testing.assert_allclose(data.weights, data_loaded1.weights)
        np.testing.assert_allclose(data.weights, data_loaded2.weights)
    with pytest.raises(
        TypeError, match="The object you are trying to serialize contains numpy arrays."
    ):
        data.to_json()
    with pytest.raises(
        TypeError, match="The object you are trying to serialize contains numpy arrays."
    ):
        data.to_yaml()
    with pytest.raises(
        TypeError, match="The object you are trying to serialize contains numpy arrays."
    ):
        data_loaded2.to_json()
    with pytest.raises(
        TypeError, match="The object you are trying to serialize contains numpy arrays."
    ):
        data_loaded2.to_yaml()



"""This  module defines the ``BasePdf`` that can be used to inherit from in order to build a custom PDF.

The ``BasePDF`` implements already a lot of ready-to-use functionality like integral, automatic normalization
and sampling.

Defining your own pdf
---------------------

A simple example:
>>> import zfit
>>> import zfit.z.numpy as znp
>>>
>>> class MyGauss(BasePDF):
>>>     def __init__(self, mean, stddev, name="MyGauss"):
>>>         super().__init__(mean=mean, stddev=stddev, name=name)
>>>
>>>     def _unnormalized_pdf(self, x):
>>>         return znp.exp((x - mean) ** 2 / (2 * stddev**2))

Notice that *here* we only specify the *function* and no normalization. This
**No** attempt to **explicitly** normalize the function should be done inside ``_unnormalized_pdf``.
The normalization is handled with another method depending on the normalization range specified.
(It *is* possible, though discouraged, to directly provide the *normalized probability* by overriding _pdf(), but
there are other, more convenient ways to add improvements like providing an analytical integrals.)

Before we create an instance, we need to create the variables to initialize it
>>> mean = zfit.Parameter("mean1", 2., 0.1, 4.2)  # signature as in RooFit: *name, initial, lower, upper*
>>> stddev = zfit.Parameter("stddev1", 5., 0.3, 10.)
Let's create an instance and some example data
>>> gauss = MyGauss(mean=mean, stddev=stddev)
>>> example_data = np.random.random(10)
Now we can get the probability
>>> probs = gauss.pdf(example_data)  # ``norm`` specifies over which range to normalize
Or the integral
>>> integral = gauss.integrate(limits=(-5, 3.1),norm=False)  # norm_range is False -> return unnormalized
integral
Or directly sample from it
>>> sample = gauss.sample(n_draws=1000, limits=(-10, 10))  # draw 1000 samples within (-10, 10)

We can create an extended PDF, which will result in anything using a ``norm_range`` to not return the
probability but the number probability (the function will be normalized to ``yield`` instead of 1 inside
the ``norm_range``)
>>> yield1 = Parameter("yield1", 100, 0, 1000)
>>> gauss_extended = gauss.create_extended(yield1)
>>> gauss.is_extended
True

>>> integral_extended = gauss.ext_integrate(limits=(-10, 10),norm=(-10, 10))  # yields approx 100

For more advanced methods and ways to register analytic integrals or overwrite certain methods, see
also the advanced models in `zfit models <https://github.com/zfit/zfit-tutorials>`_
"""
#  Copyright (c) 2023 zfit

from __future__ import annotations

from typing import TYPE_CHECKING

from tensorflow.python.util.deprecation import deprecated_args

from ..util.ztyping import ExtendedInputType, NormInputType

if TYPE_CHECKING:
    import zfit

import warnings
from contextlib import suppress

import tensorflow as tf

import zfit.z.numpy as znp
from zfit import z
from .basemodel import BaseModel
from .baseobject import extract_filter_params
from .interfaces import ZfitParameter, ZfitPDF
from .parameter import Parameter, convert_to_parameter
from .sample import extended_sampling
from .space import Space
from ..settings import run, ztypes
from ..util import ztyping
from ..util.cache import invalidate_graph
from ..util.deprecation import deprecated, deprecated_norm_range
from ..util.exception import (
    AlreadyExtendedPDFError,
    BreakingAPIChangeError,
    FunctionNotImplemented,
    NotExtendedPDFError,
    NormNotImplemented,
    SpecificFunctionNotImplemented,
)
from ..util.temporary import TemporarilySet

_BasePDF_USER_IMPL_METHODS_TO_CHECK = {}


def _BasePDF_register_check_support(has_support: bool):
    """Marks a method that the subclass either *has* to or *can't* use the ``@supports`` decorator.

    Args:
        has_support: If True, flags that it **requires** the ``@supports`` decorator. If False,
            flags that the ``@supports`` decorator is **not allowed**.
    """
    if not isinstance(has_support, bool):
        raise TypeError("Has to be boolean.")

    def register(func):
        """Register a method to be checked to (if True) *has* ``support`` or (if False) has *no* ``support``.

        Args:
            func:

        Returns:
            Function:
        """
        name = func.__name__
        _BasePDF_USER_IMPL_METHODS_TO_CHECK[name] = has_support
        func.__wrapped__ = _BasePDF_register_check_support
        return func

    return register


class BasePDF(ZfitPDF, BaseModel):
    def __init__(
        self,
        obs: ztyping.ObsTypeInput,
        params: dict[str, ZfitParameter] = None,
        dtype: type = ztypes.float,
        name: str = "BasePDF",
        extended: ExtendedInputType = None,
        norm: NormInputType = None,
        **kwargs,
    ):
        super().__init__(obs=obs, dtype=dtype, name=name, params=params, **kwargs)

        self._yield = None
        self._norm = norm
        if extended is not False and extended is not None:
            self._set_yield(extended)

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls._subclass_check_support(
            methods_to_check=_BasePDF_USER_IMPL_METHODS_TO_CHECK,
            wrapper_not_overwritten=_BasePDF_register_check_support,
        )

    def _check_input_norm(self, norm, none_is_error=False):
        if norm is None:
            norm = self.norm
        return super()._check_input_norm(norm=norm, none_is_error=none_is_error)

    def _check_input_params(self, *params):
        return tuple(convert_to_parameter(p) for p in params)

    def _func_to_integrate(self, x: ztyping.XType):
        return self.pdf(x, norm=False)

    def _func_to_sample_from(self, x):
        return self.pdf(x, norm=False)

    @property
    @deprecated(None, "Use the `norm` attribute instead.")
    def norm_range(self) -> Space | None | bool:
        """Return the current normalization range. If None and the ``obs`` have limits, they are returned.

        Returns:
            The current normalization range.
        """
        return self.norm

    @property
    def norm(self) -> Space | None | bool:
        """Return the current normalization range. If None and the ``obs`` have limits, they are returned.

        Returns:
            The current normalization range.
        """
        norm = self._norm
        if norm is None:
            norm = self.space
        return norm

    @invalidate_graph
    @deprecated(None, "Prefer to create a new PDF with `norm` set.")
    def set_norm_range(self, norm: ztyping.LimitsTypeInput):
        """Set the normalization range (temporarily if used with contextmanager).

        Args:
            norm:
        """
        norm = self._check_input_norm(norm)

        def setter(value):
            self._norm = value

        def getter():
            return self._norm

        return TemporarilySet(value=norm, setter=setter, getter=getter)

    @_BasePDF_register_check_support(True)
    def _normalization(self, norm, options):
        raise SpecificFunctionNotImplemented

    @deprecated_args(None, "Use `norm` instead.", "limits")
    def normalization(
        self,
        norm: ztyping.LimitsType,
        *,
        options=None,
        limits: ztyping.LimitsType = None,
    ) -> ztyping.XType:
        """Return the normalization of the function (usually the integral over ``norm``).

        Args:
            norm:  |@doc:pdf.param.norm| Normalization of the function.
               By default, this is the ``norm`` of the PDF (which by default is the same as
               the space of the PDF). Should be ``ZfitSpace`` to define the space
               to normalize over. |@docend:pdf.param.norm|
            options (): |@doc:pdf.param.options||@docend:pdf.param.options|

        Returns:
            The normalization value
        """
        del limits
        if options is None:
            options = {}  # TODO: pass options through
        norm = self._check_input_norm(norm, none_is_error=True)

        return self._single_hook_normalization(norm=norm, options=options)

    def _single_hook_normalization(self, norm, options):  # TODO(Mayou36): add yield?
        return self._hook_normalization(norm=norm, options=options)

    def _hook_normalization(self, norm, options):
        return self._call_normalization(norm=norm, options=options)  # no _norm_* needed

    def _call_normalization(self, norm, options):
        # TODO: caching? alternative
        with suppress(FunctionNotImplemented):
            return self._normalization(norm=norm, options=options)
        return self._fallback_normalization(norm, options=options)

    def _fallback_normalization(self, norm, options):
        return self._hook_integrate(limits=norm, norm=False, options=options)

    def _unnormalized_pdf(self, x):
        raise SpecificFunctionNotImplemented

    @deprecated(None, "Use `pdf(norm=False)` instead")
    def unnormalized_pdf(self, x: ztyping.XType) -> ztyping.XType:
        """PDF "unnormalized". Use ``functions`` for unnormalized pdfs. this is only for performance in special cases.

        Args:
            x: |@doc:pdf.param.x| Data to evaluate the method on. Should be ``ZfitData``
               or a mapping of *obs* to numpy-like arrays.
               If an array is given, the first dimension is interpreted as the events while
               the second is meant to be the dimensionality of a single event. |@docend:pdf.param.x|
        Returns:
            1-dimensional :py:class:`tf.Tensor` containing the unnormalized pdf.
        """
        with self._convert_sort_x(x) as x:
            return self._single_hook_unnormalized_pdf(x)

    def _single_hook_unnormalized_pdf(self, x):
        return self._call_unnormalized_pdf(x=x)

    def _call_unnormalized_pdf(self, x):
        # try:
        return self._unnormalized_pdf(x)

    @z.function(wraps="model")
    @deprecated_norm_range
    def ext_pdf(
        self,
        x: ztyping.XTypeInput,
        norm: ztyping.LimitsTypeInput = None,
        *,
        norm_range=None,
    ) -> ztyping.XType:
        """Probability density function scaled by yield, normalized over ``norm_range``.

        Args:
          x: |@doc:pdf.param.x| Data to evaluate the method on. Should be ``ZfitData``
               or a mapping of *obs* to numpy-like arrays.
               If an array is given, the first dimension is interpreted as the events while
               the second is meant to be the dimensionality of a single event. |@docend:pdf.param.x|
          norm: |@doc:pdf.param.norm| Normalization of the function.
               By default, this is the ``norm`` of the PDF (which by default is the same as
               the space of the PDF). Should be ``ZfitSpace`` to define the space
               to normalize over. |@docend:pdf.param.norm|

        Returns:
          :py:class:`tf.Tensor` of type `self.dtype`.
        """
        del norm_range  # taken care of in the deprecation decorator
        norm = self._check_input_norm(norm, none_is_error=True)
        if not self.is_extended:
            raise NotExtendedPDFError(f"{self} is not extended, cannot call `ext_pdf`")
        with self._convert_sort_x(x) as x:
            return self._call_ext_pdf(x, norm)

    def _call_ext_pdf(self, x, norm):
        with suppress(SpecificFunctionNotImplemented):
            return self._auto_ext_pdf(x, norm)

        # fallback
        return self.pdf(x=x, norm=norm) * self.get_yield()

    def _auto_ext_pdf(self, x, norm):
        try:
            probs = self._ext_pdf(x, norm)
        except NormNotImplemented:
            unnorm_probs = self._ext_pdf(x, False)
            normalization = self.normalization(norm)
            probs = unnorm_probs / normalization
        return probs

    @_BasePDF_register_check_support(True)
    def _ext_pdf(self, x, norm, *, norm_range=None):
        raise SpecificFunctionNotImplemented  # TODO: implement properly

    @z.function(wraps="model")
    @deprecated_norm_range
    def ext_log_pdf(
        self,
        x: ztyping.XTypeInput,
        norm: ztyping.LimitsTypeInput = None,
        *,
        norm_range=None,
    ) -> ztyping.XType:
        """Log of probability density function scaled by yield, normalized over ``norm_range``.

        Args:
          x: |@doc:pdf.param.x| Data to evaluate the method on. Should be ``ZfitData``
               or a mapping of *obs* to numpy-like arrays.
               If an array is given, the first dimension is interpreted as the events while
               the second is meant to be the dimensionality of a single event. |@docend:pdf.param.x|
          norm: |@doc:pdf.param.norm| Normalization of the function.
               By default, this is the ``norm`` of the PDF (which by default is the same as
               the space of the PDF). Should be ``ZfitSpace`` to define the space
               to normalize over. |@docend:pdf.param.norm|

        Returns:
          :py:class:`tf.Tensor` of type `self.dtype`.
        """
        assert norm_range is None
        del norm_range  # taken care of in the deprecation decorator
        norm = self._check_input_norm(norm, none_is_error=True)
        if not self.is_extended:
            raise NotExtendedPDFError(f"{self} is not extended, cannot call `ext_pdf`")
        with self._convert_sort_x(x) as x:
            return self._call_ext_log_pdf(x, norm)

    def _call_ext_log_pdf(self, x, norm):
        with suppress(SpecificFunctionNotImplemented):
            return self._auto_ext_log_pdf(x, norm)

        # fallback
        return self.log_pdf(x=x, norm=norm) + znp.log(self.get_yield())

    def _auto_ext_log_pdf(self, x, norm):
        try:
            pdf = self._ext_log_pdf(x, norm)
        except NormNotImplemented:
            unnormed_pdf = self._ext_log_pdf(x, False)
            normalization = znp.log(self.normalization(norm))
            pdf = unnormed_pdf - normalization
        return pdf

    @_BasePDF_register_check_support(True)
    def _ext_log_pdf(self, x, norm):
        raise SpecificFunctionNotImplemented

    @_BasePDF_register_check_support(True)
    def _pdf(self, x, norm, *, norm_range=None):
        raise SpecificFunctionNotImplemented

    @deprecated_norm_range
    @z.function(wraps="model")
    def pdf(
        self,
        x: ztyping.XTypeInput,
        norm: ztyping.LimitsTypeInput = None,
        *,
        norm_range=None,
    ) -> ztyping.XType:
        """Probability density function of ``x``, normalized over ``norm``.

        Args:
          x: |@doc:pdf.param.x| Data to evaluate the method on. Should be ``ZfitData``
               or a mapping of *obs* to numpy-like arrays.
               If an array is given, the first dimension is interpreted as the events while
               the second is meant to be the dimensionality of a single event. |@docend:pdf.param.x|
          norm: |@doc:pdf.param.norm| Normalization of the function.
               By default, this is the ``norm`` of the PDF (which by default is the same as
               the space of the PDF). Should be ``ZfitSpace`` to define the space
               to normalize over. |@docend:pdf.param.norm|

        Returns:
          :py:class:`tf.Tensor` of type `self.dtype`.
        """
        assert norm_range is None
        del norm_range  # taken care of in the deprecation decorator
        norm = self._check_input_norm(norm, none_is_error=True)
        with self._convert_sort_x(x) as x:
            value = self._single_hook_pdf(x=x, norm=norm)
            if run.numeric_checks:
                z.check_numerics(
                    value, message="Check if pdf output contains any NaNs of Infs"
                )
            return znp.asarray(z.to_real(value))

    def _single_hook_pdf(self, x, norm):
        return self._hook_pdf(x=x, norm=norm)

    def _hook_pdf(self, x, norm):
        return self._norm_pdf(x=x, norm=norm)

    def _norm_pdf(self, x, norm):
        return self._call_pdf(x=x, norm=norm)

    def _call_pdf(self, x, norm):
        with suppress(FunctionNotImplemented):
            return self._pdf(x, norm)
        with suppress(FunctionNotImplemented):
            return znp.exp(self._log_pdf(x, norm))
        if self.is_extended:
            with suppress(FunctionNotImplemented):
                return (
                    self._ext_pdf(x, norm) / self.get_yield()
                )  # TODO: extend/refactor the calling

        return self._fallback_pdf(x, norm)

    def _fallback_pdf(self, x, norm):
        pdf = self._call_unnormalized_pdf(x)
        if norm.has_limits:
            pdf /= self._hook_normalization(norm=norm, options={})
        return pdf

    @_BasePDF_register_check_support(False)
    @deprecated_norm_range
    def _log_pdf(self, x, norm):
        raise SpecificFunctionNotImplemented

    @deprecated_norm_range
    def log_pdf(
        self, x: ztyping.XType, norm: ztyping.LimitsType = None, *, norm_range=None
    ) -> ztyping.XType:
        """Log probability density function normalized over ``norm_range``.

        Args:
          x: |@doc:pdf.param.x| Data to evaluate the method on. Should be ``ZfitData``
               or a mapping of *obs* to numpy-like arrays.
               If an array is given, the first dimension is interpreted as the events while
               the second is meant to be the dimensionality of a single event. |@docend:pdf.param.x|
          norm: |@doc:pdf.param.norm| Normalization of the function.
               By default, this is the ``norm`` of the PDF (which by default is the same as
               the space of the PDF). Should be ``ZfitSpace`` to define the space
               to normalize over. |@docend:pdf.param.norm|

        Returns:
          A ``Tensor`` of type ``self.dtype``.
        """
        assert norm_range is None
        del norm_range  # taken care of in the deprecation decorator
        norm = self._check_input_norm(norm)
        with self._convert_sort_x(x) as x:
            return znp.asarray(z.to_real(self._single_hook_log_pdf(x=x, norm=norm)))

    def _single_hook_log_pdf(self, x, norm):
        return self._hook_log_pdf(x=x, norm=norm)

    def _hook_log_pdf(self, x, norm):
        log_prob = self._norm_log_pdf(x=x, norm=norm)
        return log_prob

    def _norm_log_pdf(self, x, norm):
        return self._call_log_pdf(x=x, norm=norm)

    def _call_log_pdf(self, x, norm):
        with suppress(FunctionNotImplemented):
            return self._log_pdf(x, norm)
        with suppress(FunctionNotImplemented):
            return znp.log(self._pdf(x, norm))
        return self._fallback_log_pdf(x, norm)

    def _fallback_log_pdf(self, x, norm):
        return znp.log(self._hook_pdf(x=x, norm=norm))

    @z.function(wraps="model")
    @deprecated_norm_range
    def ext_integrate(
        self,
        limits: ztyping.LimitsType,
        norm: ztyping.LimitsType = None,
        *,
        norm_range=None,
        options=None,
    ) -> ztyping.XType:
        """Integrate the function over ``limits`` (normalized over ``norm_range`` if not False).

        Args:
            limits: |@doc:pdf.integrate.limits| Limits of the integration. |@docend:pdf.integrate.limits|
            norm: |@doc:pdf.integrate.norm| Normalization of the integration.
               By default, this is the same as the default space of the PDF.
               ``False`` means no normalization and returns the unnormed integral. |@docend:pdf.integrate.norm|
            options: |@doc:pdf.integrate.options| Options for the integration.
               Additional options for the integration. Currently supported options are:
               - type: one of (``bins``)
                 This hints that bins are integrated. A method that is vectorizable, non-dynamic and
                 therefore less suitable for complicated functions is chosen. |@docend:pdf.integrate.options|

        Returns:
            The integral value as a scalar with shape ()
        """
        if options is None:
            options = {}
        assert norm_range is None
        del norm_range  # taken care of in the deprecation decorator
        norm = self._check_input_norm(norm)
        limits = self._check_input_limits(limits=limits)
        if not self.is_extended:
            raise NotExtendedPDFError(f"{self} is not extended, cannot call `ext_pdf`")
        return (
            self.integrate(limits=limits, norm=norm, options=options) * self.get_yield()
        )

    def _apply_yield(
        self, value: float, norm: ztyping.LimitsType, log: bool
    ) -> float | tf.Tensor:
        if self.is_extended and not norm.limits_are_false:
            if log:
                value += znp.log(self.get_yield())
            else:
                value *= self.get_yield()
        return value

    @deprecated(None, "Use the public `set_yield` instead.")
    def _set_yield_inplace(self, value: ZfitParameter | float | None):
        """Make the model extended by setting a yield.

        This does not alter the general behavior of the PDF. If there is a
        ``norm_range`` given, the output of the above functions does not represent a normalized
        probability density function anymore but corresponds to a number probability.

        Args:
            value:
        """
        self._set_yield(value=value)

    def create_extended(
        self,
        yield_: ztyping.ParamTypeInput,
        name: str = None,
        *,
        name_addition: str = None,
    ) -> ZfitPDF:
        """Return an extended version of this pdf with yield ``yield_``. The parameters are shared.

        Args:
            yield_: |@doc:pdf.param.yield| Yield (expected number of events) of the PDF.
               This is the expected number of events.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.param.yield|
            name: New name of the PDF. If ``None``, the name of the PDF with a trailing "_ext" is used.

        Returns:
            :py:class:`~zfit.core.interfaces.ZfitPDF`: a new PDF that is extended
        """
        # TODO(Mayou36): fix copy
        if name_addition is not None:
            raise BreakingAPIChangeError(
                "name_addition is not supported anymore, use `name` instead."
            )
        from zfit.models.functor import ProductPDF

        name = f"{self.name}_ext" if name is None else name

        if isinstance(self, ProductPDF):
            warnings.warn(
                "As `copy` is not yet properly implemented, this may fails (for ProductPDF for example?). This"
                "will be fixed in the future."
            )
        if self.is_extended:
            raise AlreadyExtendedPDFError(
                "This PDF is already extended, cannot create an extended one."
            )
        try:
            new_pdf = self.copy(name=name)
        except Exception as error:
            raise RuntimeError(
                f"PDF {self} could not be copied, therefore `create_extended` failed and a new "
                f"extended PDF cannot be created. As an alternative, you can use `set_yield`"
                f" to set the yield on the current PDF *inplace* (this won't return a new PDF but"
                f" instead modify the existing)."
            ) from error
        new_pdf.set_yield(value=yield_)
        return new_pdf

    def set_yield(self, value):
        """Make the model extended **inplace** by setting a yield. If possible, prefer to use ``create_extended``.

        This does not alter the general behavior of the PDF. The ``pdf`` and ``integrate`` and similar methods will
        continue to return the same - normalized to 1 - values. However, not only can this parameter be accessed
        via ``get_yield``, the methods ``ext_pdf`` and ``ext_integral`` provide a version of ``pdf`` and ``integrate``
        respecetively that is multiplied by the yield.

        These can be useful for plotting and for binned likelihoods.

        Args:
            value: |@doc:pdf.param.yield| Yield (expected number of events) of the PDF.
               This is the expected number of events.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.param.yield|
        """
        self._set_yield(value=value)

    def _set_yield(self, value: ztyping.ParamTypeInput):
        if value is None:
            raise BreakingAPIChangeError("Cannot unset a yield (anymore).")
        if self.is_extended:
            raise AlreadyExtendedPDFError(f"Cannot extend {self}, is already extended.")
        value = convert_to_parameter(value)
        self.add_cache_deps(value)
        self._yield = value

    @property
    def is_extended(self) -> bool:
        """Flag to tell whether the model is extended or not.

        Returns:
            A boolean.
        """
        return self._yield is not None

    def _hook_sample(self, limits, n):
        if n is None and self.is_extended:
            n = "extended"
        if isinstance(n, str) and n == "extended":
            if not self.is_extended:
                raise NotExtendedPDFError(
                    "Cannot use 'extended' as value for `n` on a non-extended pdf."
                )
            samples = extended_sampling(pdfs=self, limits=limits)
        elif isinstance(n, str):
            raise ValueError(
                "`n` is a string and not 'extended'. Other options are currently not implemented."
            )
        elif n is None:
            raise tf.errors.InvalidArgumentError(
                "`n` cannot be `None` if pdf is not extended."
            )
        else:
            samples = super()._hook_sample(limits=limits, n=n)
        return samples

    def get_yield(self) -> Parameter | None:
        """Return the yield (only for extended models).

        Returns:
            The yield of the current model or None
        """
        return self._yield

    @property
    def extended(self) -> Parameter | None:
        """Return the yield (only for extended models).

        Returns:
            The yield of the current model or None
        """
        return self.get_yield()

    def _get_params(
        self,
        floating: bool | None = True,
        is_yield: bool | None = None,
        extract_independent: bool | None = True,
    ) -> set[ZfitParameter]:
        params = super()._get_params(
            floating, is_yield=is_yield, extract_independent=extract_independent
        )

        if is_yield is not False:
            if self.is_extended:
                yield_params = extract_filter_params(
                    self.get_yield(),
                    floating=floating,
                    extract_independent=extract_independent,
                )
                yield_params.update(params)  # putting the yields at the beginning
                params = yield_params
            elif is_yield is True:
                raise NotExtendedPDFError(
                    "PDF is not extended but only yield parameters were requested."
                )
        return params

    def create_projection_pdf(
        self, limits: ztyping.LimitsTypeInput, *, options=None, limits_to_integrate=None
    ) -> ZfitPDF:
        """Create a PDF projection by integrating out some dimensions.

        The new projection pdf is still fully dependent on the pdf it was created with.

        Args:
            limits: |@doc:pdf.partial_integrate.limits||@docend:pdf.partial_integrate.limit|
            options: |@doc:pdf.integrate.options| Options for the integration.
               Additional options for the integration. Currently supported options are:
               - type: one of (``bins``)
                 This hints that bins are integrated. A method that is vectorizable, non-dynamic and
                 therefore less suitable for complicated functions is chosen. |@docend:pdf.integrate.options|

        Returns:
            A pdf without the dimensions from ``limits``.
        """
        if limits_to_integrate is not None:
            raise BreakingAPIChangeError(
                "Use `limits` instead of `limits_to_integrate`."
            )
        from ..models.special import SimpleFunctorPDF

        if limits_to_integrate is not None:
            limits = limits_to_integrate

        def partial_integrate_wrapped(self_simple, x):
            return self.partial_integrate(x, limits=limits, options=options)

        new_pdf = SimpleFunctorPDF(
            obs=self.space.get_subspace(
                obs=[obs for obs in self.obs if obs not in limits.obs]
            ),
            pdfs=(self,),
            func=partial_integrate_wrapped,
        )
        return new_pdf

    def copy(self, **override_parameters) -> BasePDF:
        """Creates a copy of the model.

        Note: the copy model may continue to depend on the original
        initialization arguments.

        Args:
          **override_parameters: String/value dictionary of initialization
            arguments to override with new value.

        Returns:
          A new instance of `type(self)` initialized from the union
            of self.parameters and override_parameters, i.e.,
            `dict(self.parameters, **override_parameters)`.
        """
        obs = self.norm

        # HACK(Mayou36): remove once copy is proper implemented
        from ..models.dist_tfp import WrapDistribution
        from ..models.kde import GaussianKDE1DimV1
        from ..models.polynomials import RecursivePolynomial

        if (
            type(self) == WrapDistribution
        ):  # NOT isinstance! Because e.g. Gauss wraps that and takes different args
            parameters = dict(
                distribution=self._distribution, dist_params=self.dist_params
            )
        else:
            # HACK END

            parameters = dict(self.params)
            lam = parameters.pop("lambda", None)
            if lam is not None:
                parameters["lam"] = lam

        if type(self) == GaussianKDE1DimV1:
            raise RuntimeError(
                "Cannot copy `GaussianKDE1DimV1` (yet). If you tried to make it extended, use "
                "`set_yield`"
                " instead and set it inplace."
            )
            parameters["data"] = self._original_data

        # HACK(Mayou36): copy the polynomial correct, replace 'c_0' with coeff0/coeff_0 or similar
        if isinstance(self, RecursivePolynomial):
            parameters["coeff0"] = parameters.pop("c_0", None)
            coeffs = []
            i_coeff = 1
            # collect coeffs and convert to 'coeff' list
            while True:
                coeff_name = f"c_{i_coeff}"
                try:
                    coeff = parameters.pop(coeff_name)
                except KeyError:
                    break
                else:
                    coeffs.append(coeff)
                i_coeff += 1
            parameters["coeffs"] = coeffs

        from zfit.models.functor import BaseFunctor, SumPDF

        if isinstance(self, BaseFunctor):
            parameters = {}
            if isinstance(self, SumPDF):
                fracs = self.fracs
                if not self.is_extended:
                    fracs = fracs[:-1]
                parameters.update(fracs=fracs)
            parameters.update(pdfs=self.pdfs)

        parameters.update(obs=obs, name=self.name)
        parameters.update(**override_parameters)
        # if hasattr(self, "distribution"):
        #     parameters.update(distribution=self.distribution)
        yield_ = parameters.pop("yield", None)
        new_instance = type(self)(**parameters)
        if yield_ is not None:
            new_instance.set_yield(yield_)
        return new_instance

    @deprecated_norm_range
    def as_func(self, norm: ztyping.LimitsType = False, *, norm_range=None):
        """Return a `Function` with the function `model(x, norm=norm)`.

        Args:
            norm:
        """
        assert norm_range is None
        del norm_range  # taken care of in the deprecation decorator
        from .operations import convert_pdf_to_func  # prevent circular import

        return convert_pdf_to_func(pdf=self, norm=norm)

    def __str__(self):
        return (
            "zfit.model.{type_name}("
            '"{self_name}"'
            ", dtype={dtype})".format(
                type_name=type(self).__name__,
                self_name=self.name,
                dtype=self.dtype.name,
            )
        )

    def to_unbinned(self):
        """Convert to unbinned pdf, returns self if already unbinned."""
        return self

    def to_binned(self, space, *, extended=None, norm=None):
        """Convert to binned pdf, returns self if already binned."""
        from ..models.tobinned import BinnedFromUnbinnedPDF

        return BinnedFromUnbinnedPDF(
            pdf=self, space=space, extended=extended, norm=norm
        )



#  Copyright (c) 2023 zfit
import pytest

import zfit


def test_get_dependents_is_deterministic():
    parameters = [zfit.Parameter(f"param{i}", i) for i in range(4)]
    obs = zfit.Space("obs1", (-3, 2))

    def create_pdf(params):
        param1, param2, param3, param4 = params
        gauss1 = zfit.pdf.Gauss(param1, param2, obs=obs)
        gauss2 = zfit.pdf.Gauss(param2, param2, obs=obs)
        gauss3 = zfit.pdf.Gauss(param2, param4, obs=obs)
        gauss4 = zfit.pdf.Gauss(param3, param1, obs=obs)
        sum_pdf = zfit.pdf.SumPDF(
            [gauss1, gauss2, gauss3, gauss4], fracs=[0.1, 0.3, 0.4]
        )
        return sum_pdf

    for params in (parameters, reversed(parameters)):
        pdf = create_pdf(params)
        assert (
            pdf.get_cache_deps() == pdf.get_cache_deps()
        ), "get_dependents() not deterministic"


def test_get_params():
    obs = zfit.Space("obs", (-4, 5))
    mu_nofloat = zfit.Parameter("mu_nofloat", 1, floating=False)
    mu2 = zfit.Parameter("mu2", 1)
    sigma2 = zfit.Parameter("sigma2", 2)
    sigma_comp = zfit.ComposedParameter("sigma_comp", lambda s: s * 0.7, params=sigma2)

    yield1 = zfit.Parameter("yield1", 10)
    yield2 = zfit.Parameter("yield2", 200)
    yield2_comp = zfit.ComposedParameter(
        "yield2_comp", lambda y: y * 0.9, params=yield2
    )

    gauss = zfit.pdf.Gauss(mu_nofloat, sigma_comp, obs)
    gauss2 = zfit.pdf.Gauss(mu2, sigma2, obs)
    gauss_ext = gauss.create_extended(yield1)
    gauss2_ext = gauss2.create_extended(yield2_comp)

    frac = zfit.Parameter("frac", 0.4)
    sum1 = zfit.pdf.SumPDF([gauss, gauss2], fracs=frac)
    sum1_ext = zfit.pdf.SumPDF([gauss_ext, gauss2_ext])

    assert set(gauss.get_params()) == {sigma2}

    with pytest.raises(ValueError):
        set(gauss.get_params(floating=True, is_yield=False, extract_independent=False))

    with pytest.raises(ValueError):
        set(
            gauss_ext.get_params(
                floating=True, is_yield=False, extract_independent=False
            )
        )
    assert set(
        gauss_ext.get_params(floating=None, is_yield=False, extract_independent=False)
    ) == {mu_nofloat, sigma_comp}
    with pytest.raises(ValueError):
        set(
            gauss_ext.get_params(
                floating=False, is_yield=False, extract_independent=False
            )
        )

    with pytest.raises(ValueError):
        set(
            gauss_ext.get_params(
                floating=True, is_yield=None, extract_independent=False
            )
        )

    assert set(
        gauss_ext.get_params(floating=None, is_yield=None, extract_independent=False)
    ) == {mu_nofloat, sigma_comp, yield1}
    with pytest.raises(ValueError):
        set(
            gauss_ext.get_params(
                floating=False, is_yield=None, extract_independent=False
            )
        )

    assert (
        set(
            gauss_ext.get_params(
                floating=False, is_yield=True, extract_independent=False
            )
        )
        == set()
    )
    assert set(
        gauss_ext.get_params(floating=None, is_yield=True, extract_independent=False)
    ) == {yield1}
    assert set(
        gauss_ext.get_params(floating=True, is_yield=True, extract_independent=False)
    ) == {yield1}

    # with extract deps
    assert set(
        gauss_ext.get_params(floating=True, is_yield=False, extract_independent=True)
    ) == {sigma2}
    assert set(
        gauss_ext.get_params(floating=None, is_yield=False, extract_independent=True)
    ) == {mu_nofloat, sigma2}
    assert set(
        gauss_ext.get_params(floating=False, is_yield=False, extract_independent=True)
    ) == {mu_nofloat}

    assert set(
        gauss_ext.get_params(floating=True, is_yield=None, extract_independent=True)
    ) == {sigma2, yield1}
    assert set(
        gauss_ext.get_params(floating=None, is_yield=None, extract_independent=True)
    ) == {mu_nofloat, sigma2, yield1}
    assert set(
        gauss_ext.get_params(floating=False, is_yield=None, extract_independent=True)
    ) == {mu_nofloat}

    assert set(
        gauss_ext.get_params(floating=True, is_yield=True, extract_independent=True)
    ) == {yield1}
    assert set(
        gauss_ext.get_params(floating=None, is_yield=True, extract_independent=True)
    ) == {yield1}
    assert (
        set(
            gauss_ext.get_params(
                floating=False, is_yield=True, extract_independent=True
            )
        )
        == set()
    )

    # Gauss ext2
    assert set(
        gauss2_ext.get_params(floating=True, is_yield=False, extract_independent=False)
    ) == {mu2, sigma2}
    assert set(
        gauss2_ext.get_params(floating=None, is_yield=False, extract_independent=False)
    ) == {mu2, sigma2}
    assert (
        set(
            gauss2_ext.get_params(
                floating=False, is_yield=False, extract_independent=False
            )
        )
        == set()
    )

    with pytest.raises(ValueError):
        set(
            gauss2_ext.get_params(
                floating=True, is_yield=None, extract_independent=False
            )
        )
    assert set(
        gauss2_ext.get_params(floating=None, is_yield=None, extract_independent=False)
    ) == {mu2, sigma2, yield2_comp}
    with pytest.raises(ValueError):
        set(
            gauss2_ext.get_params(
                floating=False, is_yield=None, extract_independent=False
            )
        )

    with pytest.raises(ValueError):
        set(
            gauss2_ext.get_params(
                floating=False, is_yield=True, extract_independent=False
            )
        )
    assert set(
        gauss2_ext.get_params(floating=None, is_yield=True, extract_independent=False)
    ) == {yield2_comp}
    with pytest.raises(ValueError):
        set(
            gauss2_ext.get_params(
                floating=True, is_yield=True, extract_independent=False
            )
        )

    # with extract deps
    assert set(
        gauss2_ext.get_params(floating=True, is_yield=False, extract_independent=True)
    ) == {mu2, sigma2}
    assert set(
        gauss2_ext.get_params(floating=None, is_yield=False, extract_independent=True)
    ) == {mu2, sigma2}
    yield2.floating = False
    assert (
        set(
            gauss2_ext.get_params(
                floating=False, is_yield=False, extract_independent=True
            )
        )
        == set()
    )
    yield2.floating = True
    assert (
        set(
            gauss2_ext.get_params(
                floating=False, is_yield=False, extract_independent=True
            )
        )
        == set()
    )

    assert set(
        gauss2_ext.get_params(floating=True, is_yield=None, extract_independent=True)
    ) == {mu2, sigma2, yield2}
    assert set(
        gauss2_ext.get_params(floating=None, is_yield=None, extract_independent=True)
    ) == {mu2, sigma2, yield2}
    assert (
        set(
            gauss2_ext.get_params(
                floating=False, is_yield=None, extract_independent=True
            )
        )
        == set()
    )

    assert set(
        gauss2_ext.get_params(floating=True, is_yield=True, extract_independent=True)
    ) == {yield2}
    assert set(
        gauss2_ext.get_params(floating=None, is_yield=True, extract_independent=True)
    ) == {yield2}
    assert (
        set(
            gauss2_ext.get_params(
                floating=False, is_yield=True, extract_independent=True
            )
        )
        == set()
    )

    # sum extended
    with pytest.raises(ValueError):
        assert set(
            sum1_ext.get_params(
                floating=True, is_yield=False, extract_independent=False
            )
        )
    frac0 = sum1_ext.fracs[0]
    frac1 = sum1_ext.fracs[1]
    assert set(
        sum1_ext.get_params(floating=None, is_yield=False, extract_independent=False)
    ) == {mu_nofloat, sigma_comp, mu2, sigma2, frac0, frac1}
    with pytest.raises(ValueError):
        assert set(
            sum1_ext.get_params(
                floating=False, is_yield=False, extract_independent=False
            )
        )

    with pytest.raises(ValueError):
        set(
            sum1_ext.get_params(floating=True, is_yield=None, extract_independent=False)
        )
    assert set(
        sum1_ext.get_params(floating=None, is_yield=None, extract_independent=False)
    ) == {mu_nofloat, sigma_comp, mu2, sigma2, frac0, frac1, sum1_ext.get_yield()}
    with pytest.raises(ValueError):
        set(
            sum1_ext.get_params(
                floating=False, is_yield=None, extract_independent=False
            )
        )

    with pytest.raises(ValueError):
        set(
            sum1_ext.get_params(
                floating=False, is_yield=True, extract_independent=False
            )
        )
    assert set(
        sum1_ext.get_params(floating=None, is_yield=True, extract_independent=False)
    ) == {sum1_ext.get_yield()}
    with pytest.raises(ValueError):
        set(
            sum1_ext.get_params(floating=True, is_yield=True, extract_independent=False)
        )

    # with extract deps
    assert set(
        sum1_ext.get_params(floating=True, is_yield=False, extract_independent=True)
    ) == {
        mu2,
        sigma2,
        yield1,
        yield2,  # fracs depend on them
    }
    assert set(
        sum1_ext.get_params(floating=None, is_yield=False, extract_independent=True)
    ) == {mu_nofloat, mu2, sigma2, yield1, yield2}
    assert set(
        sum1_ext.get_params(floating=False, is_yield=False, extract_independent=True)
    ) == {mu_nofloat}

    assert set(
        sum1_ext.get_params(floating=True, is_yield=None, extract_independent=True)
    ) == {mu2, sigma2, yield1, yield2, yield1, yield2}
    assert set(
        sum1_ext.get_params(floating=None, is_yield=None, extract_independent=True)
    ) == {mu_nofloat, mu2, sigma2, yield1, yield2}
    assert set(
        sum1_ext.get_params(floating=False, is_yield=None, extract_independent=True)
    ) == {mu_nofloat}
    yield1.floating = False
    assert set(
        sum1_ext.get_params(floating=False, is_yield=None, extract_independent=True)
    ) == {mu_nofloat, yield1}
    yield1.floating = True

    assert set(
        sum1_ext.get_params(floating=True, is_yield=True, extract_independent=True)
    ) == {yield1, yield2}
    assert set(
        sum1_ext.get_params(floating=None, is_yield=True, extract_independent=True)
    ) == {yield1, yield2}
    assert (
        set(
            sum1_ext.get_params(floating=False, is_yield=True, extract_independent=True)
        )
        == set()
    )



"""Baseclass for most objects appearing in zfit."""

#  Copyright (c) 2023 zfit

from __future__ import annotations

import contextlib
import itertools
from collections import OrderedDict
from collections.abc import Iterable

import tensorflow as tf
from ordered_set import OrderedSet

from .dependents import BaseDependentsMixin
from .interfaces import (
    ZfitIndependentParameter,
    ZfitNumericParametrized,
    ZfitObject,
    ZfitParameter,
    ZfitParametrized,
)
from ..util import ztyping
from ..util.cache import GraphCachable
from ..util.checks import NotSpecified
from ..util.container import convert_to_container
from ..util.exception import BreakingAPIChangeError


class BaseObject(ZfitObject):
    def __init__(self, name, **kwargs):
        assert (
            not kwargs
        ), f"kwargs not empty, the following arguments are not captured: {kwargs}"
        super().__init__()

        self._name = name  # TODO: uniquify name?

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls._repr = {}  # TODO: make repr more sophisticated

    @property
    def name(self) -> str:
        """The name of the object."""
        return self._name

    def copy(
        self, deep: bool = False, name: str = None, **overwrite_params
    ) -> ZfitObject:
        new_object = self._copy(deep=deep, name=name, overwrite_params=overwrite_params)
        return new_object

    def _copy(self, deep, name, overwrite_params):  # TODO(Mayou36) L: representation?
        raise NotImplementedError("This copy should not be used.")

    def __eq__(self, other: object) -> bool:
        if not isinstance(self, type(other)):
            return False
        with contextlib.suppress(AttributeError):
            for key, own_element in self._repr.items():
                if not own_element == other._repr.get(key):
                    return False
        return self is other
        # return True  # no break occurred

    def __hash__(self):
        return object.__hash__(self)


class BaseParametrized(BaseObject, ZfitParametrized):
    def __init__(self, params, **kwargs) -> None:
        super().__init__(**kwargs)
        from zfit.core.parameter import convert_to_parameter

        params = params or OrderedDict()
        # params = OrderedDict(sorted((n, convert_to_parameter(p)) for n, p in params.items()))
        params = {n: convert_to_parameter(p) for n, p in params.items()}  # why sorted?

        # parameters = OrderedDict(sorted(parameters))  # to always have a consistent order
        self._params = params
        self._repr["params"] = self.params

    def get_params(
        self,
        floating: bool | None = True,
        is_yield: bool | None = None,
        extract_independent: bool | None = True,
        only_floating=NotSpecified,
    ) -> set[ZfitParameter]:
        """Recursively collect parameters that this object depends on according to the filter criteria.

        Which parameters should be included can be steered using the arguments as a filter.
         - **None**: do not filter on this. E.g. ``floating=None`` will return parameters that are floating as well as
            parameters that are fixed.
         - **True**: only return parameters that fulfil this criterion
         - **False**: only return parameters that do not fulfil this criterion. E.g. ``floating=False`` will return
            only parameters that are not floating.

        Args:
            floating: if a parameter is floating, e.g. if :py:meth:`~ZfitParameter.floating` returns `True`
            is_yield: if a parameter is a yield of the _current_ model. This won't be applied recursively, but may include
               yields if they do also represent a parameter parametrizing the shape. So if the yield of the current
               model depends on other yields (or also non-yields), this will be included. If, however, just submodels
               depend on a yield (as their yield) and it is not correlated to the output of our model, they won't be
               included.
            extract_independent: If the parameter is an independent parameter, i.e. if it is a ``ZfitIndependentParameter``.
        """
        if only_floating is not NotSpecified:
            raise BreakingAPIChangeError(
                "The argument `only_floating` has been renamed to `floating`."
            )
        return self._get_params(
            floating=floating,
            is_yield=is_yield,
            extract_independent=extract_independent,
        )

    def _get_params(
        self,
        floating: bool | None,
        is_yield: bool | None,
        extract_independent: bool | None,
    ) -> set[ZfitParameter]:
        if (
            is_yield is True
        ):  # we want exclusively yields, we don't have them by default
            params = OrderedSet()
        else:
            params = self.params.values()
            params = extract_filter_params(
                params, floating=floating, extract_independent=extract_independent
            )
        return params

    @property
    def params(self) -> ztyping.ParameterType:
        return self._params


class BaseNumeric(
    GraphCachable,
    BaseDependentsMixin,
    BaseParametrized,
    ZfitNumericParametrized,
    BaseObject,
):
    def __init__(self, **kwargs):
        if "dtype" in kwargs:  # TODO(Mayou36): proper dtype handling?
            self._dtype = kwargs.pop("dtype")
        super().__init__(**kwargs)
        self.add_cache_deps(self.params.values())

    @property
    def dtype(self) -> tf.DType:
        """The dtype of the object."""
        return self._dtype

    @staticmethod
    def _filter_floating_params(params):
        params = [
            param
            for param in params
            if isinstance(param, ZfitIndependentParameter) and param.floating
        ]
        return params


def extract_filter_params(
    params: Iterable[ZfitParametrized],
    floating: bool | None = True,
    extract_independent: bool | None = True,
) -> set[ZfitParameter]:
    params = convert_to_container(params, container=OrderedSet)

    if extract_independent is not False:
        params_indep = OrderedSet(
            itertools.chain.from_iterable(
                param.get_params(
                    floating=floating, extract_independent=True, is_yield=None
                )
                for param in params
            )
        )
        if extract_independent is True:
            params = params_indep
        else:  # None
            params |= params_indep
    if floating is not None:
        if not extract_independent and not all(param.independent for param in params):
            raise ValueError(
                "Since `extract_dependent` is not set to True, there are maybe dependent parameters for "
                "which `floating` is an ill-defined attribute."
            )
        params = OrderedSet(p for p in params if p.floating == floating)
    return params



#  Copyright (c) 2023 zfit

from ._loss.binnedloss import (
    ExtendedBinnedNLL,
    BinnedNLL,
    ExtendedBinnedChi2,
    BinnedChi2,
)
from .core.loss import BaseLoss, ExtendedUnbinnedNLL, SimpleLoss, UnbinnedNLL

__all__ = [
    "ExtendedUnbinnedNLL",
    "UnbinnedNLL",
    "BinnedNLL",
    "ExtendedBinnedNLL",
    "BaseLoss",
    "SimpleLoss",
    "ExtendedBinnedNLL",
    "BinnedChi2",
    "ExtendedBinnedChi2",
]



"""Used to make pytest functions available globally."""
#  Copyright (c) 2023 zfit

import collections.abc
import os
import pathlib
import sys

import dill
import matplotlib.pyplot as plt
import numpy as np
import pytest

try:
    import pytest_randomly
except ImportError:
    pass
else:
    import zfit

    pytest_randomly.random_seeder = [zfit.settings.set_seed]
init_modules = sys.modules.keys()


@pytest.fixture(autouse=True)
def setup_teardown():
    import zfit

    old_chunksize = zfit.run.chunking.max_n_points
    old_active = zfit.run.chunking.active
    old_graph_mode = zfit.run.get_graph_mode()
    old_autograd_mode = zfit.run.get_autograd_mode()

    for m in sys.modules.keys():
        if m not in init_modules:
            del sys.modules[m]

    yield

    from zfit.core.parameter import ZfitParameterMixin, _reset_auto_number

    ZfitParameterMixin._existing_params.clear()
    _reset_auto_number()

    from zfit.util.cache import clear_graph_cache

    clear_graph_cache()
    import zfit

    zfit.run.chunking.active = old_active
    zfit.run.chunking.max_n_points = old_chunksize
    zfit.run.set_graph_mode(old_graph_mode)
    zfit.run.set_autograd_mode(old_autograd_mode)
    zfit.run.set_graph_cache_size()
    for m in sys.modules.keys():
        if m not in init_modules:
            del sys.modules[m]

    zfit.settings.set_seed(None)
    import gc

    gc.collect()


def pytest_addoption(parser):
    parser.addoption("--longtests", action="store_true", default=False)
    parser.addoption("--longtests-kde", action="store_true", default=False)
    parser.addoption("--recreate-truth", action="store_true", default=False)


def pytest_configure():
    here = os.path.dirname(os.path.abspath(__file__))
    images_dir = pathlib.Path(here).joinpath(
        "..", "docs", "images", "_generated_by_tests"
    )
    images_dir.mkdir(exist_ok=True)

    def savefig(figure=None, folder=None):
        if figure is None:
            figure = plt.gcf()
        title_sanitized = (
            figure.axes[0]
            .get_title()
            .replace(" ", "_")
            .replace("$", "_")
            .replace("\\", "_")
            .replace("__", "_")
        )
        title_sanitized = (
            title_sanitized.replace("/", "_")
            .replace(".", "_")
            .replace(":", "_")
            .replace(",", "")
        )
        if not title_sanitized:
            raise RuntimeError("Title has to be set for plot that should be saved.")
        foldersave = images_dir
        if folder is not None:
            foldersave = foldersave.joinpath(folder)
        foldersave.mkdir(exist_ok=True, parents=True)
        savepath = foldersave.joinpath(title_sanitized)
        plt.savefig(str(savepath))

    pytest.zfit_savefig = savefig


ARBITRARY_VALUE = "ZFIT_ARBITRARY_VALUE"


def cleanup_recursive(dict1, dict2):
    """Compare two dicts recursively."""
    if not (
        isinstance(dict1, collections.abc.Mapping)
        and isinstance(dict2, collections.abc.Mapping)
    ):
        return dict1, dict2
    dict1, dict2 = dict1.copy(), dict2.copy()
    for key in set(dict1.keys()) | set(dict2.keys()):
        val1 = dict1.get(key)
        val2 = dict2.get(key)
        if isinstance(val1, str) and val1 == ARBITRARY_VALUE:
            dict2[key] = ARBITRARY_VALUE
        elif isinstance(val2, str) and val2 == ARBITRARY_VALUE:
            dict1[key] = ARBITRARY_VALUE
        elif isinstance(val1, collections.abc.Mapping) or isinstance(
            val2, collections.abc.Mapping
        ):
            dict1sub, dict2sub = cleanup_recursive(dict1.get(key), dict2.get(key))
            if dict1sub is not None:
                dict1[key] = dict1sub
            if dict2sub is not None:
                dict2[key] = dict2sub
        elif isinstance(val1, np.ndarray):
            dict1[key] = val1.tolist()
        elif isinstance(val2, np.ndarray):
            dict2[key] = val2.tolist()
        elif key == "value_fn":  # We have a composed parameter
            if isinstance(val1, str):
                dict1[key] = "FUNC_IGNORED_NOT_STABLE"
            if isinstance(val2, str):
                dict2[key] = "FUNC_IGNORED_NOT_STABLE"
    return dict1, dict2


@pytest.helpers.register
def cleanup_hs3(obj1, obj2):
    """Compare two HS3 dicts."""

    if not isinstance(obj1, collections.abc.Mapping) or not isinstance(
        obj2, collections.abc.Mapping
    ):
        raise TypeError(
            f"obj1 and obj2 both need to be of type 'Mapping', are {obj1} and {obj2}"
        )
    missing2 = set(obj1.keys()) - set(obj2.keys())
    missing1 = set(obj2.keys()) - set(obj1.keys())
    if missing1 and missing2:
        raise ValueError(
            f"Both objects are missing keys: {missing1} and {missing2}. "
            f"obj1: {obj1}, obj2: {obj2}"
        )
    return cleanup_recursive(obj1, obj2)


@pytest.helpers.register
def get_truth(folder, filename, request, newval=None):
    """Get the truth value for a given folder and filename.

    Args:
        folder (str): The folder in which the file is located.
        filename (str): The filename of the file.
        newval (Any): If given, the truth value will be overwritten with this value.

    Returns:
        The truth value.
    """
    if newval is None:
        raise ValueError(
            "New value has to be given. This will only be stored and overwrite the current file *if* the --recreate-truth flag is given. "
            "Otherwise, the truth value will be loaded from the file."
        )
    current_dir = pathlib.Path(__file__).parent
    static_dir = current_dir / "truth" / folder
    static_dir.mkdir(exist_ok=True, parents=True)

    filepath = static_dir / (filename)
    # check if need to update value first
    recreate_truth = request.config.getoption("--recreate-truth")
    if recreate_truth:
        if filepath.suffix == ".json":
            import json

            if isinstance(newval, str):
                newval = json.loads(newval)
            with open(filepath, "w") as f:
                json.dump(newval, f)
        elif filepath.suffix == ".yaml":
            import yaml

            with open(filepath, "w") as f:
                yaml.dump(newval, f)

        elif filepath.suffix == ".asdf":
            newval.write_to(filepath)
        else:
            raise ValueError(
                f"Filetype {filepath.suffix} not supported. Needs manual implementation of the truth value in get_truth function."
            )

    if not filepath.exists():
        raise FileNotFoundError(f"File {filepath} does not exist")
    if filepath.suffix == ".json":
        import json

        with open(filepath, "r") as f:
            return json.load(f)
    elif filepath.suffix == ".yaml":
        import yaml

        with open(filepath, "r") as f:
            return yaml.safe_load(f)
    elif filepath.suffix == ".asdf":
        import asdf

        with asdf.open(filepath, copy_arrays=True) as f:
            return asdf.AsdfFile(f.tree, copy_arrays=True)
    else:
        raise ValueError(f"Filetype {filepath.suffix} not supported")
    raise ValueError(f"Folder {folder} not supported.")



#  Copyright (c) 2024 zfit
import itertools

import numpy as np
import pytest
import tensorflow_probability as tfp
from matplotlib import pyplot as plt

import zfit
from zfit import z
from zfit.core.interfaces import ZfitParameter


@pytest.mark.skip()  # copy not yet implemented
def test_copy_kde():
    size = 500
    data = np.random.normal(size=size, loc=2, scale=3)

    limits = (-15, 5)
    obs = zfit.Space("obs1", limits=limits)
    kde_adaptive = zfit.pdf.GaussianKDE1DimV1(
        data=data, bandwidth="adaptive", obs=obs, truncate=False
    )
    kde_adaptive.copy()


def create_pdf_sample(npoints=500, upper=None):
    if upper is None:
        upper = 11
    limits = (-13, upper)
    obs = zfit.Space("obs1", limits=limits)
    cb = zfit.pdf.CrystalBall(mu=2, sigma=3, alpha=1, n=25, obs=obs)
    gauss = zfit.pdf.Gauss(mu=-5, sigma=2.5, obs=obs)
    pdf = zfit.pdf.SumPDF([cb, gauss], fracs=0.8)
    data = pdf.sample(n=npoints)
    return data, obs, pdf


def test_grid_kde():
    data, obs, pdf = create_pdf_sample()
    with pytest.raises(ValueError):
        zfit.pdf.KDE1DimGrid(obs=obs, data=data, bandwidth="eoaue")


def test_exact_kde():
    data, obs, pdf = create_pdf_sample()
    with pytest.raises(ValueError):
        zfit.pdf.KDE1DimExact(obs=obs, data=data, bandwidth="eoaue")


def create_kde(
    kdetype=None,
    npoints=1500,
    cfgonly=False,
    nonly=False,
    full=True,
    upper=None,
    legacy=True,
    padding=False,
):
    import tensorflow as tf

    import zfit
    import zfit.z.numpy as znp

    npoints_lim = min(npoints, 2500)

    class StudentT(tfp.distributions.StudentT):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, df=5, **kwargs)

    configs = []
    with tf.init_scope():
        if cfgonly or nonly:
            hparam1 = 0.9
        else:
            hparam1 = zfit.Parameter("h1", 0.9)
    comb = itertools.product(
        [
            ("bandwidth", h)
            for h in [
                None,
                hparam1,
                "adaptive",
                "silverman",
                "scott",
                np.random.uniform(0.1, 1.5, size=npoints_lim),
            ]
        ],
        [("truncate", bo) for bo in [False, True]],
        [("type", zfit.pdf.GaussianKDE1DimV1)],
        [("npoints", npoints_lim)],
    )
    if full or legacy:
        if legacy:
            comb = [next(comb)]
        configs.extend(comb)

    with tf.init_scope():
        if cfgonly or nonly:
            hparam2 = 0.9
        else:
            hparam2 = zfit.Parameter("h2", 0.92)
    comb = itertools.product(
        [
            ("bandwidth", h)
            for h in [
                None,
                hparam2,
                "adaptive_zfit",
                "adaptive_geom",
                "adaptive_std",
                "silverman",
                "scott",
                "isj",
                np.random.uniform(0.1, 1.5, size=npoints_lim),
            ]
        ],
        [("weights", weight) for weight in [None, znp.ones(shape=npoints_lim)]],
        [("padding", padding)],
        [("kernel", dist) for dist in [None, StudentT]],
        [("type", zfit.pdf.KDE1DimExact)],
        [("npoints", npoints_lim)],
    )
    if not full:
        comb = [next(comb)]
    configs.extend(comb)

    # Grid PDFs

    num_grid_size_maybe = 800
    comb = itertools.product(
        [
            ("bandwidth", h)
            for h in [
                "adaptive_geom",
                None,
                hparam2,
                "silverman",
                "scott",
                "adaptive_zfit",
                np.random.uniform(0.1, 1.5, size=num_grid_size_maybe),
            ]
        ],
        [("weights", weight) for weight in [None, znp.ones(shape=npoints)]],
        [("padding", padding)],
        [("kernel", dist) for dist in [None, StudentT]],
        [("type", zfit.pdf.KDE1DimGrid)],
        [("num_grid_points", n) for n in [None, num_grid_size_maybe]],
        [("binning_method", method) for method in ["linear", "simple"]],
    )
    comb = [
        c
        for c in comb
        if not isinstance(c[0][1], np.ndarray) or c[5][1] == num_grid_size_maybe
    ]

    if not full:
        comb = [comb[0]]
    configs.extend(comb)

    # FFT combinations

    comb = itertools.product(
        [("bandwidth", h) for h in [None, "scott"]],  # TODO: adaptive?
        [("weights", weight) for weight in [None, znp.ones(shape=npoints)]],
        [("padding", padding)],
        [("kernel", dist) for dist in [None, StudentT]],
        [("type", zfit.pdf.KDE1DimFFT)],
        [("num_grid_points", n) for n in [None, num_grid_size_maybe]],
        [("binning_method", method) for method in ["linear", "simple"]],
    )

    if not full:
        comb = [next(comb)]
    configs.extend(comb)

    # IFJ combinations

    comb = itertools.product(
        [("weights", weight) for weight in [None, znp.ones(shape=npoints)]],
        [("padding", padding)],
        [("type", zfit.pdf.KDE1DimISJ)],
        [("num_grid_points", n) for n in [None, num_grid_size_maybe]],
        [("binning_method", method) for method in ["linear", "simple"]],
    )

    if not full:
        comb = [next(comb)]
    configs.extend(comb)

    # end config builder
    if nonly:
        return len(configs)
    cfg = configs[kdetype]
    cfg = dict(cfg)
    if cfgonly:
        return cfg
    _ = cfg.pop("npoints", None)

    constructor = cfg.pop("type")
    if constructor in (zfit.pdf.GaussianKDE1DimV1, zfit.pdf.KDE1DimExact):
        npoints = npoints_lim

    data, obs, pdf = create_pdf_sample(npoints, upper=upper)

    cfg["data"] = data
    cfg["obs"] = obs

    kde = constructor(**cfg)

    return kde, pdf, data.value()[:, 0]


def _get_print_kde(**kwargs):
    kdes = create_kde(**kwargs)
    return kdes


@pytest.hookimpl
def pytest_generate_tests(metafunc):
    if metafunc.function.__name__ == "test_all_kde":
        full = metafunc.config.getoption("--longtests")
        full = full or metafunc.config.getoption("--longtests-kde")

        metafunc.parametrize(
            "kdetype", [i for i in range(_get_print_kde(nonly=True, full=full))]
        )


# @pytest.mark.flaky(3)
@pytest.mark.parametrize(
    "jit",
    [
        False,
        # True
    ],
)
@pytest.mark.parametrize("npoints", [1100, 500_000])
def test_all_kde(kdetype, npoints, jit, request):
    import zfit

    full = request.config.getoption("--longtests")
    full = full or request.config.getoption("--longtests-kde")
    cfg = create_kde(kdetype=kdetype, npoints=npoints, cfgonly=True, full=full)
    kdetype = kdetype, npoints
    if jit:
        run_jit = z.function(_run)
        (
            expected_integral,
            integral,
            name,
            prob,
            prob_true,
            rel_tol,
            sample,
            sample2,
            x,
            name,
            data,
        ) = run_jit(kdetype, full=full)
    else:
        (
            expected_integral,
            integral,
            name,
            prob,
            prob_true,
            rel_tol,
            sample,
            sample2,
            x,
            name,
            data,
        ) = _run(kdetype, full=full)

    expected_integral = zfit.run(expected_integral)

    if not jit:
        import matplotlib.pyplot as plt

        plt.figure()
        kernel_used = cfg.get("kernel")
        if kernel_used is not None:
            kernel_print = f", kernel={kernel_used.__name__}"
        else:
            kernel_print = ""
        bandwidth_printready = cfg.get("bandwidth")
        if isinstance(bandwidth_printready, ZfitParameter):
            bandwidth_printready = f"Param({float(bandwidth_printready.value())}"
        if bandwidth_printready is None:
            bandwidth_print = ""
        else:
            bandwidth_print = f", h={bandwidth_printready}"

        num_grid_points = cfg.get("num_grid_points")
        if num_grid_points is not None:
            grid_print = f", grid={num_grid_points}"
        else:
            grid_print = ""
        binning_method = cfg.get("binning_method")
        if binning_method is not None:
            binning_print = f", binning={binning_method}"
        else:
            binning_print = ""
        weights = cfg.get("weights")
        if weights is not None:
            weights_print = ", weighted"
        else:
            weights_print = ""
        npoints_plot = cfg.get("npoints", kdetype[1])
        plt.title(
            f"{name} with {npoints_plot} points{weights_print}{bandwidth_print}{kernel_print}{grid_print}{binning_print}",
            fontsize=10,
        )
        plt.plot(x, prob, label="KDE")
        plt.plot(x, prob_true, label="true PDF")
        data_np = zfit.run(data)
        plt.hist(data_np, bins=40, density=True, alpha=0.3, label="Kernel points")
        plt.legend()

    abs_tol = 0.005 if kdetype[1] > 3000 else 0.03
    tolfac = 6 if not cfg["type"] == tfp.distributions.Normal else 1
    if cfg.get("binning_method") == "simple":
        tolfac *= 6
    assert zfit.run(integral) == pytest.approx(expected_integral, abs=abs_tol * tolfac)
    assert tuple(sample.shape) == (1, 1)
    assert tuple(sample2.shape) == (1500, 1)
    assert prob.shape.rank == 1
    assert prob.shape[0] == x.shape[0]
    assert np.mean(prob - prob_true) < 0.07 * tolfac
    # make sure that on average, most values are close
    assert np.mean(
        (prob / prob_true)[prob_true > np.mean(prob_true)] ** 2
    ) == pytest.approx(1, abs=0.1 * tolfac)
    rtol = 0.05
    np.testing.assert_allclose(prob, prob_true, rtol=rtol, atol=0.01 * tolfac)


def _run(kdetype, full, upper=None, legacy=True, padding=False):
    from zfit.z import numpy as znp

    kde, pdf, data = create_kde(
        *kdetype, full=full, upper=upper, legacy=legacy, padding=padding
    )
    integral = kde.integrate(
        limits=kde.space,
        norm=(-3, 2),
    )
    expected_integral = kde.integrate(
        limits=kde.space,
        norm=(-3, 2),
    )
    rel_tol = 0.04
    sample = kde.sample(1).value()
    sample2 = kde.sample(1500).value()
    x = znp.linspace(*kde.space.limit1d, 30000)
    prob = kde.pdf(x)

    prob_true = pdf.pdf(x)

    return (
        expected_integral,
        integral,
        kde.name,
        prob,
        prob_true,
        rel_tol,
        sample,
        sample2,
        x,
        kde.name,
        data,
    )


@pytest.mark.parametrize(
    "kdetype", [i for i in range(_get_print_kde(nonly=True, full=False, legacy=False))]
)
@pytest.mark.parametrize("npoints", [1100, 500_000])
@pytest.mark.parametrize("upper", [-4, -1, 3])
def test_kde_border(kdetype, npoints, upper):
    import zfit

    cfg = create_kde(
        kdetype=kdetype, npoints=npoints, cfgonly=True, full=False, legacy=False
    )
    kdetype = kdetype, npoints

    (
        expected_integral,
        integral,
        name,
        prob,
        prob_true,
        rel_tol,
        sample,
        sample2,
        x,
        name,
        data,
    ) = _run(kdetype, full=False, upper=upper, legacy=False, padding=True)

    plt.figure()
    if (kernel_used := cfg.get("kernel")) is not None:
        kernel_print = f", kernel={kernel_used.__name__}"
    else:
        kernel_print = ""
    bandwidth_printready = cfg.get("bandwidth")
    if isinstance(bandwidth_printready, ZfitParameter):
        bandwidth_printready = f"Param({float(bandwidth_printready.value())}"
    if bandwidth_printready is None:
        bandwidth_print = ""
    else:
        bandwidth_print = f", h={bandwidth_printready}"

    if (num_grid_points := cfg.get("num_grid_points")) is not None:
        grid_print = f", grid={num_grid_points}"
    else:
        grid_print = ""
    if (binning_method := cfg.get("binning_method")) is not None:
        binning_print = f", binning={binning_method}"
    else:
        binning_print = ""
    weights = cfg.get("weights")
    if weights is not None:
        weights_print = ", weighted"
    else:
        weights_print = ""
    npoints_plot = cfg.get("npoints", kdetype[1])
    plt.title(
        f"{name}, {npoints_plot} points, upper={upper}{weights_print}{bandwidth_print}"
        f"{kernel_print}{grid_print}{binning_print}",
        fontsize=10,
    )
    plt.plot(x, prob, label="KDE")
    plt.plot(x, prob_true, label="true PDF")
    data_np = zfit.run(data)
    plt.hist(data_np, bins=40, density=True, alpha=0.3, label="Kernel points")
    plt.legend()

    integral = zfit.run(integral)
    expected_integral = zfit.run(expected_integral)
    abs_tol = 0.05
    assert zfit.run(integral) == pytest.approx(expected_integral, abs=abs_tol)
    assert tuple(sample.shape) == (1, 1)
    assert tuple(sample2.shape) == (1500, 1)
    assert prob.shape.rank == 1
    assert np.mean(prob - prob_true) < 0.07
    # make sure that on average, most values are close
    assert np.mean(
        (prob / prob_true)[prob_true > np.mean(prob_true)] ** 2
    ) == pytest.approx(1, abs=0.15)
    rtol = 0.05
    np.testing.assert_allclose(prob, prob_true, rtol=rtol, atol=0.07)



"""This example illustrates how to minimize an arbitrary Python function using the zfit minimizer.

This may has some overhead in the beginning and won't be instantly fast compared to other libraries if run once.

Copyright (c) 2021 zfit
"""

#  Copyright (c) 2022 zfit

import numpy as np

import zfit

# set everything to numpy mode. This is only needed if we
# don't use
zfit.run.set_autograd_mode(False)
zfit.run.set_graph_mode(False)

# create our favorite minimizer
minimizer = zfit.minimize.IpyoptV1()


# minimizer = zfit.minimize.Minuit()
# minimizer = zfit.minimize.ScipyTrustConstrV1()
# minimizer = zfit.minimize.NLoptLBFGSV1()


def func(x):
    x = np.asarray(x)  # make sure it's an array
    return np.sum((x - 0.1) ** 2 + x[1] ** 4)


# We can use the same in pure TF, then we can also use
# the analytic gradient
#
# import tensorflow as tf
#
# @tf.function
# def func(x):
#     x = tf.convert_to_tensor(x)  # make sure it's an array
#     return tf.reduce_sum((x - 0.1) ** 2 + x[1] ** 4)


# we can also use a more complicated function instead
# from scipy.optimize import rosen as func


# we need to set the errordef, the definition of "1 sigma"
func.errordef = 0.5

# initial parameters
params = [1, -3, 2, 1.4, 11]
# or for a more fine-grained control
# params = {
#     'value': [1, -3, 2, 1.4, 11],  # mandatory
#     'lower': [-2, -5, -5, -10, -15],  # lower bound, can be omitted
#     'upper': [2, 4, 5, 10, 15],  # upper bound, can be omitted
#     'step_size': [0.1] * 5,  # initial step size, can be omitted
# }

# minimize
result = minimizer.minimize(func, params)

# estimate errors
result.hesse(name="hesse")
result.errors(name="errors")
print(result)



#  Copyright (c) 2023 zfit
from __future__ import annotations

from typing import Tuple, Union, Optional

from typing import Literal

from pydantic import Field, validator, root_validator

from .serializer import BaseRepr
from ..core.space import Space
from ..util.exception import WorkInProgressError

NumericTyped = Union[float, int]

NameObsTyped = Optional[Union[Tuple[str], str]]


class SpaceRepr(BaseRepr):
    _implementation = Space
    hs3_type: Literal["Space"] = Field("Space", alias="type")
    name: str
    lower: Optional[NumericTyped] = Field(alias="min")
    upper: Optional[NumericTyped] = Field(alias="max")
    binning: Optional[float] = None  # TODO: binning

    @root_validator(pre=True)
    def _validate_pre(cls, values):
        if cls.orm_mode(values):
            if values["n_obs"] > 1:
                raise RuntimeError(
                    "Multiple observables are not supported yet. For PDFs with multiple observables, "
                    "this should work. But directly dumping a multidimensional Space is not supported."
                )
            values = dict(values)
            values["name"] = values.pop("obs")[0]

        return values

    @validator("lower", pre=True)
    def _validate_lower(cls, v):
        if cls.orm_mode(v):
            v = v[0, 0]
        return v

    @validator("upper", pre=True)
    def _validate_upper(cls, v):
        if cls.orm_mode(v):
            v = v[0, 0]
        return v

    @validator("binning", pre=True, allow_reuse=True)
    def validate_binning(cls, v):
        if v is not None:
            raise WorkInProgressError("Binning is not implemented yet")
        return v

    def _to_orm(self, init) -> SpaceRepr._implementation:
        init["limits"] = init.pop("lower", None), init.pop("upper", None)
        init["obs"] = init.pop("name")
        init = super()._to_orm(init)
        return init



#  Copyright (c) 2024 zfit

from __future__ import annotations

from collections.abc import Mapping, Iterable

import tensorflow as tf
from uhi.typing.plottable import PlottableHistogram

import zfit.z.numpy as znp
from zfit import z
from zfit.core.binnedpdf import BaseBinnedPDFV1
from ..core import parameter
from ..core.interfaces import ZfitBinnedPDF
from ..util import ztyping
from ..util.exception import SpecificFunctionNotImplemented
from ..z.interpolate_spline import interpolate_spline


@z.function(wraps="tensor")
def spline_interpolator(alpha, alphas, densities):
    alphas = alphas[None, :, None]
    shape = tf.shape(densities[0])
    densities_flat = [znp.reshape(density, [-1]) for density in densities]
    densities_flat = znp.stack(densities_flat, axis=0)
    alpha_shaped = znp.reshape(alpha, [1, -1, 1])

    y_flat = interpolate_spline(
        train_points=alphas,
        train_values=densities_flat[None, ...],
        query_points=alpha_shaped,
        order=2,
    )
    y_flat = y_flat[0, 0]
    y = tf.reshape(y_flat, shape)
    return y


class SplineMorphingPDF(BaseBinnedPDFV1):
    _morphing_interpolator = staticmethod(spline_interpolator)

    def __init__(
        self,
        alpha: ztyping.ParamTypeInput,
        hists: (
            Mapping[float | int, Iterable[ZfitBinnedPDF]]
            | list[ZfitBinnedPDF]
            | tuple[ZfitBinnedPDF]
        ),
        extended: ztyping.ExtendedInputType = None,
        norm: ztyping.NormInputType = None,
    ):
        """Morphing a set of histograms with a spline interpolation.

        Args:
            alpha: Parameter for the spline interpolation.
            hists: A mapping of alpha values to histograms. This allows for arbitrary interpolation points.
                If a list or tuple of exactly three PDFs is given, this corresponds to the histograms at
                alhpa equal to -1, 0 and 1 respectively.
            extended: |@doc:pdf.init.extended| The overall yield of the PDF.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.init.extended|
            norm: |@doc:pdf.init.norm| Normalization of the PDF.
               By default, this is the same as the default space of the PDF. |@docend:pdf.init.norm|
        """
        if isinstance(hists, (list, tuple)):
            if len(hists) != 3:
                raise ValueError(
                    "If hists is a list, it is assumed to correspond to an alpha of -1, 0 and 1."
                    f" hists is {hists} and has length {len(hists)}."
                )
            else:
                hists = {
                    float(i - 1): hist for i, hist in enumerate(hists)
                }  # mapping to -1, 0, 1

        hists_clean = {}
        for a, hist in hists.items():
            if isinstance(hist, PlottableHistogram):
                from zfit.models.histogram import HistogramPDF

                hist = HistogramPDF(hist)
            if isinstance(hist, ZfitBinnedPDF):
                hists[a] = hist
            else:
                raise TypeError(
                    f"hist {hist} is not a ZfitBinnedPDF or a UHI histogram."
                )

        self.hists = hists
        self.alpha = alpha
        obs = list(hists.values())[0].space
        all_extended = all(hist.is_extended for hist in hists.values())
        if extended is None:  # TODO: yields?
            extended = all_extended
        self._automatically_extended = None
        if extended is True:  # create the yield automatically
            self._automatically_extended = True
            if not all_extended:
                raise ValueError(
                    "If extended is True, all PDFs must be extended to create the yield automatically."
                )
            alphas = znp.array(list(self.hists.keys()), dtype=znp.float64)

            def interpolated_yield(params):
                alpha = params["alpha"]
                densities = tuple(
                    params[f"{i}"] for i in range(len(params) - 1)
                )  # params has n hist entries + 1 alpha entry
                return spline_interpolator(
                    alpha=alpha, alphas=alphas, densities=densities
                )

            number = parameter.get_auto_number()
            yields = {f"{i}": hist.get_yield() for i, hist in enumerate(hists.values())}
            yields["alpha"] = alpha
            new_yield = parameter.ComposedParameter(
                f"AUTOGEN_{number}_interpolated_yield",
                interpolated_yield,
                params=yields,
            )
            extended = new_yield
        elif extended is not False:
            self._automatically_extended = False
        super().__init__(
            obs=obs,
            extended=extended,
            norm=norm,
            params={"alpha": alpha},
            name="LinearMorphing",
        )

    def _counts(self, x, norm):
        if not self._automatically_extended:
            raise SpecificFunctionNotImplemented
        densities = [hist.counts(x, norm=norm) for hist in self.hists.values()]
        alphas = znp.array(list(self.hists.keys()), dtype=znp.float64)
        alpha = self.params["alpha"].value()
        y = self._morphing_interpolator(alpha, alphas, densities)
        return y

    def _rel_counts(self, x, norm):
        densities = [hist.rel_counts(x, norm=norm) for hist in self.hists.values()]
        alphas = znp.array(list(self.hists.keys()), dtype=znp.float64)
        alpha = self.params["alpha"].value()
        y = self._morphing_interpolator(alpha, alphas, densities)
        return y

    def _ext_pdf(self, x, norm):
        if not self._automatically_extended:
            raise SpecificFunctionNotImplemented
        densities = [hist.ext_pdf(x, norm=norm) for hist in self.hists.values()]
        alphas = znp.array(list(self.hists.keys()), dtype=znp.float64)
        alpha = self.params["alpha"].value()
        y = self._morphing_interpolator(alpha, alphas, densities)
        return y

    def _pdf(self, x, norm):
        densities = [hist.pdf(x, norm=norm) for hist in self.hists.values()]
        alphas = znp.array(list(self.hists.keys()), dtype=znp.float64)
        alpha = self.params["alpha"].value()
        y = self._morphing_interpolator(alpha, alphas, densities)
        return y



#  Copyright (c) 2022 zfit
from __future__ import annotations

import pydantic
import tensorflow as tf

from zfit import z
from zfit.core.binning import unbinned_to_binindex
from zfit.core.interfaces import ZfitSpace
from zfit.core.space import supports
from zfit.models.functor import BaseFunctor
from zfit.util.ztyping import ExtendedInputType, NormInputType
from zfit.z import numpy as znp


class UnbinnedFromBinnedPDF(BaseFunctor):
    def __init__(
        self,
        pdf,
        obs=None,
        *,
        extended: ExtendedInputType = None,
        norm: NormInputType = None,
    ):
        """Create a unbinned pdf from a binned pdf.

        Args:
            pdf: Binned PDF that will be used as a step function.
            obs: |@doc:pdf.init.obs| Observables of the
               model. This will be used as the default space of the PDF and,
               if not given explicitly, as the normalization range.

               The default space is used for example in the sample method: if no
               sampling limits are given, the default space is used.

               The observables are not equal to the domain as it does not restrict or
               truncate the model outside this range. |@docend:pdf.init.obs|
            extended: |@doc:pdf.init.extended| The overall yield of the PDF.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.init.extended|
            norm: |@doc:pdf.init.norm| Normalization of the PDF.
               By default, this is the same as the default space of the PDF. |@docend:pdf.init.norm|
        """
        if extended is None:
            if extended is not False and pdf.is_extended:
                extended = pdf.get_yield()
        if obs is None:
            obs = pdf.space
            obs = obs.with_binning(None)
        super().__init__(pdfs=pdf, obs=obs, extended=extended, norm=norm)
        self._binned_space = self.pdfs[0].space.with_obs(self.space)
        self._binned_norm = self.pdfs[0].norm.with_obs(self.space)

    @supports(norm="norm", multiple_limits=True)
    def _pdf(self, x, norm):
        binned_space = self.pdfs[0].space
        binindices = unbinned_to_binindex(x, binned_space, flow=True)
        pdf = self.pdfs[0]

        binned_norm = norm if norm is False else self._binned_norm
        values = pdf.pdf(binned_space, norm=binned_norm)

        # because we have the flow, so we need to make it here with pads
        padded_values = znp.pad(
            values, znp.ones((z._get_ndims(values), 2)), mode="constant"
        )  # for overflow
        ordered_values = tf.gather_nd(padded_values, indices=binindices)
        return ordered_values

    @supports(norm="norm", multiple_limits=True)
    def _ext_pdf(self, x, norm):
        binned_space = self.pdfs[0].space
        binindices = unbinned_to_binindex(x, binned_space, flow=True)

        pdf = self.pdfs[0]

        binned_norm = norm if norm is False else self._binned_norm
        values = pdf.ext_pdf(binned_space, norm=binned_norm)
        ndim = z._get_ndims(values)

        # because we have the flow, so we need to make it here with pads
        padded_values = znp.pad(
            values, znp.ones((ndim, 2)), mode="constant"
        )  # for overflow
        ordered_values = tf.gather_nd(padded_values, indices=binindices)
        return ordered_values

    @supports(norm="norm", multiple_limits=True)
    def _integrate(self, limits, norm, options=None):
        binned_norm = norm if norm is False else self._binned_norm
        return self.pdfs[0].integrate(limits, norm=binned_norm, options=options)

    @supports(norm=True, multiple_limits=True)
    def _ext_integrate(self, limits, norm, options):
        binned_norm = norm if norm is False else self._binned_norm
        return self.pdfs[0].ext_integrate(limits, norm=binned_norm, options=options)

    @supports(norm=True, multiple_limits=True)
    def _sample(self, n, limits: ZfitSpace, *, prng=None):
        if prng is None:
            prng = z.random.get_prng()

        pdf = self.pdfs[0]
        # TODO: use real limits, currently not supported in binned sample
        sample = pdf.sample(n=n)

        edges = sample.space.binning.edges
        ndim = len(edges)
        edges = [znp.array(edge) for edge in edges]
        edges_flat = [znp.reshape(edge, [-1]) for edge in edges]
        lowers = [edge[:-1] for edge in edges_flat]
        uppers = [edge[1:] for edge in edges_flat]
        lowers_meshed = znp.meshgrid(*lowers, indexing="ij")
        uppers_meshed = znp.meshgrid(*uppers, indexing="ij")
        lowers_meshed_flat = [
            znp.reshape(lower_mesh, [-1]) for lower_mesh in lowers_meshed
        ]
        uppers_meshed_flat = [
            znp.reshape(upper_mesh, [-1]) for upper_mesh in uppers_meshed
        ]
        lower_flat = znp.stack(lowers_meshed_flat, axis=-1)
        upper_flat = znp.stack(uppers_meshed_flat, axis=-1)

        counts_flat = znp.reshape(sample.values(), (-1,))
        counts_flat = tf.cast(
            counts_flat, znp.int32
        )  # TODO: what if we have fractions?
        lower_flat_repeated = tf.repeat(lower_flat, counts_flat, axis=0)
        upper_flat_repeated = tf.repeat(upper_flat, counts_flat, axis=0)
        sample_unbinned = prng.uniform(
            (znp.sum(counts_flat), ndim),
            minval=lower_flat_repeated,
            maxval=upper_flat_repeated,
            dtype=self.dtype,
        )
        return sample_unbinned


class TypedSplinePDF(pydantic.BaseModel):
    order: pydantic.conint(ge=0)



#  Copyright (c) 2022 zfit
import hist
import mplhep
import numpy as np
from matplotlib import pyplot as plt

import zfit

# noinspection PyTypeChecker
histos = []
for i in range(5):
    h = hist.Hist(hist.axis.Regular(13, -3, 2, name="x", flow=False))
    x = np.random.normal(size=1_000_000 * (i + 1)) + i**1.5 / 2 * ((-1) ** i)
    h.fill(x=x)
    histos.append(h)
mplhep.histplot(
    histos, stack=True, histtype="fill", label=[f"process {i + 1}" for i in range(5)]
)
plt.legend()
pdfs = [zfit.pdf.HistogramPDF(h) for h in histos]
sumpdf = zfit.pdf.BinnedSumPDF(pdfs)


h_back = sumpdf.to_hist()
pdf_syst = zfit.pdf.BinwiseScaleModifier(sumpdf, modifiers=True)

mplhep.histplot(h_back)

print(h_back)
# uncomment to show plots
# plt.show()



#  Copyright (c) 2022 zfit

from __future__ import annotations

import collections
from collections.abc import Mapping

import tensorflow_probability as tfp
from zfit_interface.variables import ZfitVar

from zfit.core.interfaces import ZfitParameter, ZfitSpace, ZfitData
from zfit.util.container import convert_to_container


@tfp.experimental.auto_composite_tensor()
class ValueHolder(tfp.experimental.AutoCompositeTensor):
    def __init__(
        self,
        args,
        variables: Mapping,
        norm: ValueHolder = None,
        target=None,
        holders=None,
    ):
        args = convert_to_container(args)
        variables = self._check_input_variables(variables)
        varmap = {name: var.name for name, var in variables.items()}
        self.target = target
        self._varmap = varmap
        self._vararg = self._create_vararg_map(args, varmap)

        # needed to create auto composite
        self.norm = norm
        self.holders = holders
        self.variables = variables
        self.args = args

    def get_var(self, name):
        if name not in self._varmap:
            raise ValueError(
                f"{name} is not a valid name. Has to be one of {tuple(self._varmap.keys())}"
            )
        varname = self._varmap["name"]
        for arg in self.args:
            if isinstance(arg, ZfitVar):
                if varname == arg.name:
                    return arg
            elif isinstance(arg, ZfitData):
                if varname in arg.obs:
                    return arg[varname]
            else:
                assert (
                    False
                ), "We missed something somewhere. Please report this, it's a bug."

    def _check_input_variables(self, variables):
        if not isinstance(variables, collections.abc.Mapping):
            raise TypeError(f"variables has to be a Mapping, not {variables}")
        not_var = {var for var in variables.values() if not isinstance(var, ZfitVar)}
        if not_var:
            raise TypeError(
                f"The following values in {variables} are not ZfitVar: {not_var}"
            )

        return variables

    def __getitem__(self, item):
        try:
            index = self.names.index[item]
        except KeyError as error:
            raise KeyError(f"{self} does not contain {item}.") from error
        return self.args[index]

    def __contains__(self, item):
        return item in self.names

    @property
    def params(self):
        return {
            k: v for k, v in zip(self.names, self.args) if isinstance(v, ZfitParameter)
        }

    @property
    def space(self):
        return {k: v for k, v in zip(self.names, self.args) if isinstance(v, ZfitSpace)}

    @property
    def datasets(self):
        return {k: v for k, v in zip(self.names, self.args) if isinstance(v, ZfitData)}
